# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Library to convert a dataset from one file format to another."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping, Sequence
import dataclasses
import functools
import os
import re
from typing import Callable, Type

from etils import epy

with epy.lazy_imports():
  # pylint: disable=g-import-not-at-top
  import concurrent.futures
  import tqdm

  from absl import logging
  import apache_beam as beam
  from etils import epath
  import tensorflow as tf
  from tensorflow_datasets.core import constants
  from tensorflow_datasets.core import dataset_info as dataset_info_lib
  from tensorflow_datasets.core import file_adapters
  from tensorflow_datasets.core import naming
  from tensorflow_datasets.core import splits as splits_lib
  from tensorflow_datasets.core.proto import dataset_info_pb2
  from tensorflow_datasets.core.utils import file_utils
  from tensorflow_datasets.core.utils import py_utils
  from tensorflow_datasets.core.utils import type_utils

  # pylint: enable=g-import-not-at-top


ConvertFn = Callable[[tf.train.Example], bytes]


@dataclasses.dataclass(frozen=True)
class ConvertConfig:
  """Configuration for the conversion of a shard.

  Attributes:
    out_file_format: the format to which the dataset should be converted to.
    in_file_format: the format of the dataset to convert. If `None`, then the
      config is not yet complete.
    convert_fn: optional function to convert a single example.
    overwrite: whether to overwrite existing shards.
    fail_on_error: whether to fail on error when converting a shard or just log
      the error.
    use_beam: whether to use Beam to convert datasets. Useful for big datasets.
    num_workers: number of workers to use when not using Beam. If `use_beam` is
      set, this flag is ignored. If `num_workers=1`, the conversion will be done
      sequentially.
    in_file_adapter: file adapter for the `in_file_format`.
    out_file_adapter: file adapter for the `out_file_format`.
  """

  out_file_format: file_adapters.FileFormat
  in_file_format: file_adapters.FileFormat | None = None
  convert_fn: ConvertFn | None = None
  overwrite: bool = False
  fail_on_error: bool = False
  use_beam: bool = False
  num_workers: int = 8

  @property
  def in_file_adapter(self) -> Type[file_adapters.FileAdapter]:
    if self.in_file_format is None:
      raise ValueError('in_file_format must be set!')
    return file_adapters.ADAPTER_FOR_FORMAT[self.in_file_format]

  @property
  def out_file_adapter(self) -> Type[file_adapters.FileAdapter]:
    return file_adapters.ADAPTER_FOR_FORMAT[self.out_file_format]

  def with_in_file_format(
      self, in_file_format: str | file_adapters.FileFormat
  ) -> ConvertConfig:
    """Returns a new config with the given `in_file_format`."""
    in_file_format = file_adapters.FileFormat(in_file_format)
    return dataclasses.replace(self, in_file_format=in_file_format)


@dataclasses.dataclass(frozen=True)
class ShardInstruction:
  """Instruction for how one single shard should be converted."""

  in_path: epath.Path
  out_path: epath.Path
  config: ConvertConfig

  def convert(self) -> epath.Path | None:
    """Converts the shard to the desired file format.

    Returns:
      The path of the converted shard or `None` if the shard was not converted.

    Raises:
      Exception: if the shard conversion failed and `config.fail_on_error` is
        `True`, else logs the error.
    """

    def read_in() -> Iterator[type_utils.KeySerializedExample]:
      in_dataset = self.config.in_file_adapter.make_tf_data(
          filename=self.in_path
      )
      for i, row in tqdm.tqdm(
          enumerate(in_dataset),
          unit=' examples',
          desc=f'Shard {self.in_path.name}',
      ):
        if self.config.convert_fn is not None:
          yield i, self.config.convert_fn(row)
        else:
          yield i, row.numpy()

    try:
      with py_utils.incomplete_file(self.out_path) as tmp_file:
        self.config.out_file_adapter.write_examples(
            path=tmp_file, iterator=read_in()
        )
      return self.out_path
    except Exception as e:  # pylint: disable=broad-except
      if self.config.fail_on_error:
        raise e
      else:
        logging.exception(
            'Failed to convert shard %s (format=%s) to %s (format=%s)!',
            self.in_path,
            self.config.in_file_adapter.FILE_SUFFIX,
            self.out_path,
            self.config.out_file_adapter.FILE_SUFFIX,
        )


def _shard_instructions_for_split(
    split_info: splits_lib.SplitInfo,
    out_path: epath.Path,
    convert_config: ConvertConfig,
) -> list[ShardInstruction]:
  """Returns shard instructions for the given split."""

  if split_info.filename_template is None:
    msg = f'Filename template for split {split_info.name} is empty.'
    if convert_config.fail_on_error:
      raise ValueError(msg)
    else:
      logging.error(msg)
      return []

  in_filename_template = split_info.filename_template
  out_filename_template = in_filename_template.replace(
      data_dir=out_path,
      filetype_suffix=convert_config.out_file_format.file_suffix,
  )
  num_shards = len(split_info.shard_lengths)
  if num_shards <= 0:
    msg = f'{num_shards=} must be positive!'
    if convert_config.fail_on_error:
      raise ValueError(msg)
    else:
      logging.error(msg)
      return []

  instructions = []
  existing_files = set(out_filename_template.data_dir.glob('*'))
  for shard_index in range(num_shards):
    out_path = out_filename_template.sharded_filepath(
        shard_index=shard_index, num_shards=num_shards
    )
    if out_path in existing_files:
      logging.info('Skipping %s because it exists.', out_path)
      continue
    in_file_path = in_filename_template.sharded_filepath(
        shard_index=shard_index, num_shards=num_shards
    )
    instructions.append(
        ShardInstruction(
            in_path=in_file_path,
            out_path=out_path,
            config=convert_config,
        )
    )
  return instructions


def get_all_shard_instructions(
    info: dataset_info_pb2.DatasetInfo,
    in_path: epath.Path,
    out_path: epath.Path,
    convert_config: ConvertConfig,
) -> list[ShardInstruction]:
  """Returns all shard instructions for the given dataset info."""
  if info.file_format is None:
    msg = 'No file format was found in the dataset info!'
    if convert_config.fail_on_error:
      raise ValueError(msg)
    else:
      logging.error(msg)
      return []

  convert_config = convert_config.with_in_file_format(info.file_format)
  shard_instructions = []
  splits_dict = dataset_info_lib.get_split_dict_from_proto(
      dataset_info_proto=info,
      data_dir=in_path,
      file_format=convert_config.in_file_format,
  )
  for split_info in splits_dict.values():
    shard_instructions.extend(
        _shard_instructions_for_split(
            split_info=split_info,
            out_path=out_path,
            convert_config=convert_config,
        )
    )
  return shard_instructions


def _get_root_data_dir(
    in_dir: epath.Path, info: dataset_info_pb2.DatasetInfo
) -> epath.Path:
  in_dir = os.fspath(in_dir)
  if info.config_name:
    parts = [info.name, info.config_name, info.version]
  else:
    parts = [info.name, info.version]
  relative_data_dir = os.path.join(*parts)
  return epath.Path(re.sub(rf'{relative_data_dir}/?$', '', in_dir))


class ConvertMetadataFn(beam.DoFn):
  """Beam DoFn to convert metadata for a single dataset version."""

  def process(
      self,
      count,
      in_dir: epath.Path,
      info: dataset_info_pb2.DatasetInfo,
      out_path: epath.Path,
      convert_config: ConvertConfig,
  ):
    # This is necessary because sometimes `beam.combiners.Count.Globally()`
    # returned an integer and not a pCollection.
    if not isinstance(count, int):
      count = beam.pvalue.AsSingleton(count)
    convert_metadata(
        in_dir=in_dir,
        info=info,
        out_path=out_path,
        convert_config=convert_config,
        num_converted_shards=count,
    )


def convert_metadata(
    in_dir: epath.Path,
    info: dataset_info_pb2.DatasetInfo,
    out_path: epath.Path,
    convert_config: ConvertConfig,
    num_converted_shards: int | None = None,
) -> None:
  """Converts all metadata to the converted dataset.

  Args:
    in_dir: folder that contains the dataset to convert.
    info: dataset info of the dataset to convert.
    out_path: folder where the converted dataset should be stored.
    convert_config: configuration for the conversion.
    num_converted_shards: number of shards that were successfully converted,
      which is used to check that the conversion was successful. If part of a
      beam pipeline, this comes from `beam.combiners.Count.Globally()`.
  """
  splits_dict = dataset_info_lib.get_split_dict_from_proto(
      dataset_info_proto=info,
      data_dir=in_dir,
      file_format=convert_config.out_file_format,
  )

  missing_shards_per_split = {}
  for split_info in splits_dict.values():
    num_available_shards = len(
        split_info.get_available_shards(
            out_path, file_format=convert_config.out_file_format
        )
    )
    if num_converted_shards != num_available_shards:
      logging.warning(
          'Amount of converted shards calculated during conversion (%d) does'
          ' not match the amount of available shards in the data dir (%d) for'
          ' split %s.'
      )
    if num_available_shards < split_info.num_shards:
      missing_shards_per_split[split_info.name] = (
          num_available_shards,
          split_info.num_shards,
      )
      error_message = (
          (
              f'Found {num_available_shards} shards for split'
              f' {split_info.name}, but expected'
              f' {split_info.num_shards} shards.'
          ),
      )
      if convert_config.fail_on_error:
        raise ValueError(error_message)
      else:
        logging.error(error_message)

    elif num_available_shards > split_info.num_shards:
      msg = (
          f'Found more shards ({num_available_shards}) for split'
          f' {split_info.name}, but expected only'
          f' {split_info.num_shards} shards.'
      )
      if convert_config.fail_on_error:
        raise ValueError(msg)
      else:
        logging.error(msg)
        return

  if in_dir == out_path:
    if missing_shards_per_split:
      missing_shard_info = [
          f'split {name}: {available} / {total} shards'
          for name, (available, total) in missing_shards_per_split.items()
      ]
      logging.error(
          'Skipping updating metadata in %s because shards are missing: %s',
          os.fspath(in_dir),
          ', '.join(missing_shard_info),
      )
      return

    # File format was added to an existing dataset.
    # Add the file format to `alternative_file_formats` field.
    if convert_config.out_file_format not in info.alternative_file_formats:
      info.alternative_file_formats.append(convert_config.out_file_format.value)
      dataset_info_lib.write_dataset_info_proto(info, dataset_info_dir=out_path)
    else:
      logging.info(
          'File format %s is already an alternative file format of the dataset'
          ' in %s. Skipping updating metadata..',
          convert_config.out_file_format.value,
          os.fspath(in_dir),
      )
    return

  # Copy all json files except dataset_info.json because it needs to be updated.
  # Even if there are missing shards, it's better to copy the metadata than not
  # copying it at all.
  for json_file in in_dir.glob('*.json'):
    if json_file.name == constants.DATASET_INFO_FILENAME:
      continue
    out_file = out_path / json_file.name
    if out_file.exists():
      logging.info(
          'Not copying %s because it already exists.', os.fspath(json_file)
      )
    else:
      json_file.copy(out_file)
    logging.info('Copied %s to %s', json_file, out_file)

  # Update dataset info and store it.
  in_dataset_reference = naming.DatasetReference(
      dataset_name=info.name,
      config=info.config_name,
      version=info.version,
      data_dir=os.fspath(_get_root_data_dir(in_dir, info)),
  )
  # Record the source TFDS dataset. Note that existing data source accesses will
  # not be removed.
  dataset_info_lib.add_tfds_data_source_access(
      dataset_info_proto=info,
      dataset_reference=in_dataset_reference,
  )
  info.file_format = convert_config.out_file_format.value
  dataset_info_lib.write_dataset_info_proto(info, dataset_info_dir=out_path)


def _convert_dataset(
    info: dataset_info_pb2.DatasetInfo,
    dataset_dir: epath.Path,
    out_dir: epath.Path,
    convert_config: ConvertConfig,
    pipeline: beam.Pipeline | None = None,
) -> None:
  """Converts a single dataset version to the given file format."""
  logging.info(
      'Converting shards in %s to %s, saving in %s.',
      dataset_dir,
      convert_config.out_file_format.value,
      out_dir,
  )
  if dataset_dir != out_dir:
    if convert_config.overwrite:
      out_dir.unlink(missing_ok=True)
    out_dir.mkdir(parents=True, exist_ok=True)

  shard_instructions = get_all_shard_instructions(
      info=info,
      in_path=dataset_dir,
      out_path=out_dir,
      convert_config=convert_config,
  )

  if not shard_instructions:
    logging.warning(
        'No shard instructions found for %s. This could mean that all shards'
        ' were already converted',
        dataset_dir,
    )
    return
  else:
    logging.info('Found %d shards to convert.', len(shard_instructions))

  if pipeline is not None:
    converted_shards = (
        pipeline
        | f'CreateShardInstructions for {dataset_dir}'
        >> beam.Create(shard_instructions)
        | f'ConvertShards for {dataset_dir}'
        >> beam.Map(lambda shard_instruction: shard_instruction.convert())
        | f'Filter out shards that were not successfully converted for {dataset_dir}'
        >> beam.Filter(lambda shard_instruction: shard_instruction is not None)
    )
    count_shards = (
        converted_shards
        | f'CountConvertedShards for {dataset_dir}'
        >> beam.combiners.Count.Globally()
    )
    _ = count_shards | f'ConvertMetadata for {dataset_dir}' >> beam.ParDo(
        ConvertMetadataFn(),
        in_dir=dataset_dir,
        info=info,
        out_path=out_dir,
        convert_config=convert_config,
    )

  else:
    converted_shards = 0
    for shard_instruction in tqdm.tqdm(
        shard_instructions,
        unit=' shards',
        desc=f'Shards in {os.fspath(dataset_dir)}',
    ):
      result = shard_instruction.convert()
      if result is not None:
        converted_shards += 1
    logging.info('Converting metadata in %s.', dataset_dir)
    convert_metadata(
        in_dir=dataset_dir,
        info=info,
        out_path=out_dir,
        convert_config=convert_config,
        num_converted_shards=converted_shards,
    )


def _remove_incomplete_files(path: epath.Path) -> None:
  num_incomplete_files = 0
  for incomplete_file in path.glob(f'*{constants.INCOMPLETE_PREFIX}*'):
    if py_utils.is_incomplete_file(incomplete_file):
      incomplete_file.unlink()
      num_incomplete_files += 1
  logging.info('Removed %d incomplete files.', num_incomplete_files)


def _get_info_for_dirs_to_convert(
    from_dir: epath.Path,
    to_dir: epath.Path,
    out_file_format: file_adapters.FileFormat,
    overwrite: bool,
) -> dataset_info_pb2.DatasetInfo | None:
  """Returns the dataset info for the given dataset dirs."""
  try:
    dataset_info_proto = dataset_info_lib.read_proto_from_builder_dir(from_dir)
  except Exception:  # pylint: disable=broad-except
    logging.exception('Failed to read dataset info from %s', from_dir)
    return None
  in_file_format = file_adapters.FileFormat(dataset_info_proto.file_format)
  if out_file_format == in_file_format:
    if os.fspath(from_dir) == os.fspath(to_dir):
      logging.warning(
          'The file format to convert to (%s) is already the default file'
          ' format of the dataset in %s, and no different output folder is'
          ' specified. Skipping conversion.',
          out_file_format.value,
          os.fspath(from_dir),
      )
      return None
    else:
      logging.info(
          'The file format to convert to (%s) is the same as the default file'
          ' format, but the converted output is being written to a different'
          ' folder. The shards will be converted anyway from: %s, to: %s',
          out_file_format.value,
          os.fspath(from_dir),
          os.fspath(to_dir),
      )
      return dataset_info_proto
  if out_file_format.file_suffix in dataset_info_proto.alternative_file_formats:
    if overwrite:
      logging.warning(
          'The file format to convert to (%s) is already an alternative file'
          ' format for the dataset in %s. Overwriting the shards!',
          out_file_format.value,
          os.fspath(from_dir),
      )
    elif os.fspath(from_dir) == os.fspath(to_dir):
      logging.info(
          'The file format to convert to (%s) is already an alternative file'
          ' format of the dataset in %s. Converting missing shards if present.',
          os.fspath(from_dir),
          out_file_format.value,
      )
    else:
      logging.warning(
          'The file format to convert to (%s) is already an alternative file'
          ' format, but the converted output is being written to a different'
          ' folder, so the shards will be converted anyway. From: %s, to: %s',
          out_file_format.value,
          os.fspath(from_dir),
          os.fspath(to_dir),
      )
  else:
    logging.info(
        'The file format to convert to (%s) is not an alternative file format'
        ' of the dataset in %s. Converting the dataset.',
        out_file_format.value,
        os.fspath(from_dir),
    )
  return dataset_info_proto


def _convert_dataset_dirs(
    from_to_dirs: Mapping[epath.Path, epath.Path],
    convert_config: ConvertConfig,
) -> None:
  """Converts all datasets in the given `from_to_dirs` parameter.

  Args:
    from_to_dirs: mapping from specific input dataset folders to the folder
      where the converted dataset should be stored.
    convert_config: configuration for the conversion.
  """
  logging.info('Converting %d datasets.', len(from_to_dirs))
  found_dataset_versions: dict[epath.Path, dataset_info_pb2.DatasetInfo] = {}

  if convert_config.num_workers > 1:

    def _process_get_infos(from_to_dir):
      from_dir, to_dir = from_to_dir
      return from_dir, _get_info_for_dirs_to_convert(
          from_dir=from_dir,
          to_dir=to_dir,
          out_file_format=convert_config.out_file_format,
          overwrite=convert_config.overwrite,
      )

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=convert_config.num_workers
    ) as executor:
      for from_dir, info in executor.map(
          _process_get_infos,
          from_to_dirs.items(),
      ):
        if info is not None:
          found_dataset_versions[from_dir] = info
  else:
    for from_dir, to_dir in tqdm.tqdm(
        from_to_dirs.items(), unit=' directories'
    ):
      info = _get_info_for_dirs_to_convert(
          from_dir=from_dir,
          to_dir=to_dir,
          out_file_format=convert_config.out_file_format,
          overwrite=convert_config.overwrite,
      )
      if info is not None:
        found_dataset_versions[from_dir] = info
  convert_dataset_fn = functools.partial(
      _convert_dataset, convert_config=convert_config
  )

  # First convert all shards (with or without Beam), then convert the metadata.
  if convert_config.use_beam:
    runner = None
    with beam.Pipeline(runner=runner) as pipeline:
      for dataset_dir, info in found_dataset_versions.items():
        out_dir = from_to_dirs[dataset_dir]
        convert_dataset_fn(
            info=info,
            dataset_dir=dataset_dir,
            out_dir=out_dir,
            pipeline=pipeline,
        )
  elif convert_config.num_workers > 1:
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=convert_config.num_workers
    ) as executor:
      for dataset_dir, info in found_dataset_versions.items():
        out_dir = from_to_dirs[dataset_dir]
        executor.submit(
            convert_dataset_fn,
            info=info,
            dataset_dir=dataset_dir,
            out_dir=out_dir,
        )
  else:
    for dataset_dir, info in tqdm.tqdm(
        found_dataset_versions.items(), unit=' datasets'
    ):
      out_dir = from_to_dirs[dataset_dir]
      convert_dataset_fn(
          info=info,
          dataset_dir=dataset_dir,
          out_dir=out_dir,
      )

  logging.info(
      'All metadata and shards have been converted. Now removing incomplete'
      ' files.'
  )
  for out_dir in from_to_dirs.values():
    logging.info('Removing incomplete files in %s.', out_dir)
    _remove_incomplete_files(out_dir)


def _create_from_to_dirs(
    references: Iterable[naming.DatasetReference],
    root_in_dir: epath.Path,
    out_path: epath.Path | None,
) -> Mapping[epath.Path, epath.Path]:
  """Returns a mapping from dataset dirs to their corresponding out dirs."""
  from_to_dirs: dict[epath.Path, epath.Path] = {}
  for reference in references:
    dataset_dir = reference.dataset_dir()
    if not out_path:
      out_dir = dataset_dir
    else:
      out_dir = _create_out_dir(
          dataset_dir=dataset_dir,
          root_in_dir=root_in_dir,
          root_out_dir=out_path,
      )
    from_to_dirs[dataset_dir] = out_dir
  return from_to_dirs


def convert_root_data_dir(
    root_data_dir: epath.PathLike,
    out_dir: epath.PathLike | None,
    convert_config: ConvertConfig,
) -> None:
  """Converts all datasets found in the given dataset dir.

  Args:
    root_data_dir: folder that contains one or multiple TFDS datasets, each with
      their own configs and versions.
    out_dir: folder where the converted datasets should be stored. Datasets will
      be stored with the same folder structure as the input folder. If `None`,
      the converted datasets will be stored in the same folder as the input
      datasets.
    convert_config: configuration for the conversion.
  """
  root_data_dir = epath.Path(root_data_dir)
  out_path = epath.Path(out_dir) if out_dir is not None else None

  references = file_utils.list_datasets_in_data_dir(
      data_dir=root_data_dir,
      include_configs=True,
      include_versions=True,
      include_old_tfds_version=True,
  )
  from_to_dirs = _create_from_to_dirs(
      references=references, root_in_dir=root_data_dir, out_path=out_path
  )

  if not from_to_dirs:
    raise ValueError(f'No datasets found in the root data dir {root_data_dir}.')

  _convert_dataset_dirs(
      from_to_dirs=from_to_dirs,
      convert_config=convert_config,
  )


def _create_out_dir(
    dataset_dir: epath.Path,
    root_in_dir: epath.Path,
    root_out_dir: epath.Path,
) -> epath.Path:
  """Returns the folder where the data should be written."""
  relative_path = os.fspath(dataset_dir).removeprefix(os.fspath(root_in_dir))
  relative_path = relative_path.removeprefix('/')
  return epath.Path(root_out_dir) / relative_path


def convert_dataset_dir(
    dataset_dir: epath.PathLike,
    out_dir: epath.PathLike | None,
    convert_config: ConvertConfig,
) -> None:
  """Converts all datasets found in the given dataset dir.

  Args:
    dataset_dir: folder that contains a single dataset with all its configs and
      versions.
    out_dir: folder where the converted datasets should be stored. Datasets will
      be stored with the same folder structure as the input folder. If `None`,
      the converted shards will be stored in the same folder as the input
      datasets.
    convert_config: configuration for the conversion.
  """
  dataset_dir = epath.Path(dataset_dir)
  out_path = epath.Path(out_dir) if out_dir is not None else None

  logging.info(
      'Converting all configs and versions in dataset dir %s, saving to %s.',
      dataset_dir,
      out_dir,
  )

  references = file_utils.list_dataset_variants(
      dataset_dir=dataset_dir,
      include_old_tfds_version=True,
  )
  from_to_dirs = _create_from_to_dirs(
      references=references, root_in_dir=dataset_dir, out_path=out_path
  )

  if not from_to_dirs:
    raise ValueError(f'No datasets found in the dataset dir {dataset_dir}.')

  _convert_dataset_dirs(
      from_to_dirs=from_to_dirs,
      convert_config=convert_config,
  )


def convert_dataset(
    out_dir: epath.PathLike | None,
    out_file_format: str | file_adapters.FileFormat,
    root_data_dir: epath.PathLike | None = None,
    dataset_dir: epath.PathLike | None = None,
    dataset_version_dir: (
        epath.PathLike | Sequence[epath.PathLike] | None
    ) = None,
    overwrite: bool = False,
    use_beam: bool = False,
    num_workers: int = 8,
    convert_fn: ConvertFn | None = None,
    fail_on_error: bool = True,
) -> None:
  """Convert a dataset from one file format to another format.

  Note that exactly one of `root_data_dir`, `dataset_dir`, or
  `dataset_version_dir` must be specified.

  Args:
    out_dir: folder where the converted datasets should be stored. Datasets will
      be stored with the same folder structure as the input folder. If `None`,
      the converted shards will be stored in the same folder as the input
      datasets.
    out_file_format: file format to which the dataset should be converted.
    root_data_dir: folder that contains one or multiple TFDS datasets, each with
      their own configs and versions.
    dataset_dir: folder that contains a single dataset with all its configs and
      versions.
    dataset_version_dir: a single or list of folders that each contains a single
      dataset version. If multiple folders are specified, `out_dir` should be
      `None`, since each dataset will be converted in the same folder as the
      input dataset.
    overwrite: whether to overwrite folders in `out_dir` if they already exist.
    use_beam: whether to use Beam to convert datasets. Useful for big datasets.
    num_workers: number of workers to use when not using Beam. If `use_beam` is
      set, this flag is ignored. If `num_workers=1`, the conversion will be done
      sequentially.
    convert_fn: optional function to convert TF examples into whatever is
      desired.
    fail_on_error: whether to fail the entire pipeline when one shard fails to
      convert. When converting an entire root data dir, setting this to `True`
      will fail the entire pipeline if a single shard fails to convert.
  """
  if (
      root_data_dir is None
      and dataset_dir is None
      and dataset_version_dir is None
  ):
    raise ValueError(
        'One of `root_data_dir`, `dataset_dir`, or `dataset_version_dir`'
        ' must be specified.'
    )

  if isinstance(out_file_format, str):
    out_file_format = file_adapters.file_format_from_suffix(out_file_format)

  convert_config = ConvertConfig(
      overwrite=overwrite,
      out_file_format=out_file_format,
      in_file_format=None,  # Not yet known, will be set later.
      convert_fn=convert_fn,
      fail_on_error=fail_on_error,
      use_beam=use_beam,
      num_workers=num_workers,
  )

  if root_data_dir:
    convert_root_data_dir(
        root_data_dir=root_data_dir,
        out_dir=out_dir,
        convert_config=convert_config,
    )
  elif dataset_dir:
    convert_dataset_dir(
        dataset_dir=dataset_dir,
        out_dir=out_dir,
        convert_config=convert_config,
    )
  elif dataset_version_dir:
    if isinstance(dataset_version_dir, str):
      dataset_version_dir = [dataset_version_dir]

    if len(dataset_version_dir) > 1 and out_dir is not None:
      raise ValueError(
          'If multiple dataset version dirs are specified, `out_dir` must be'
          ' `None`, since each dataset will be converted in the same folder as'
          ' the input dataset.'
      )

    from_to_dirs = {}
    for path in dataset_version_dir:
      if out_dir is None:
        from_to_dirs[epath.Path(path)] = epath.Path(path)
      else:
        from_to_dirs[epath.Path(path)] = epath.Path(out_dir)

    _convert_dataset_dirs(
        from_to_dirs=from_to_dirs,
        convert_config=convert_config,
    )
  else:
    raise ValueError(
        'At least one of `root_data_dir`, `dataset_dir`, or'
        ' `dataset_version_dir` must be specified.'
    )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Asimov datasets."""

from tensorflow_datasets.robotics import dataset_importer_builder

ASIMOV_CITATION = """
@article{sermanet2025asimov,
  author    = {Pierre Sermanet and Anirudha Majumdar and Alex Irpan and Dmitry Kalashnikov and Vikas Sindhwani},
  title     = {Generating Robot Constitutions & Benchmarks for Semantic Safety},
  journal   = {arXiv preprint arXiv:2503.08663},
  url       = {https://arxiv.org/abs/2503.08663},
  year      = {2025},
}
"""

ASIMOV_HOMEPAGE = 'https://asimov-benchmark.github.io/'


class AsimovDilemmasAutoVal(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_dilemmas_auto_val` dataset."""

  def get_description(self):
    return (
        'Binary dilemma questions generated from counterfactual situations used'
        ' to auto-amend generated constitutions (validation set).'
    )

  def get_citation(self):
    return ASIMOV_CITATION

  def get_homepage(self):
    return ASIMOV_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_dilemmas_auto_val/0.1.0'


class AsimovDilemmasScifiTrain(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_dilemmas_scifi_train` dataset."""

  def get_description(self):
    return (
        'Multiple-choice ethical questions (with desirable and undesirable'
        ' answers) based on situations inspired from Science Fiction literature'
        ' (training set).'
    )

  def get_citation(self):
    return ASIMOV_CITATION

  def get_homepage(self):
    return ASIMOV_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_dilemmas_scifi_train/0.1.0'


class AsimovDilemmasScifiVal(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_dilemmas_scifi_val` dataset."""

  def get_description(self):
    return (
        'Multiple-choice ethical questions (with desirable and undesirable'
        ' answers) based on situations inspired from Science Fiction literature'
        ' (validation set).'
    )

  def get_citation(self):
    return ASIMOV_CITATION

  def get_homepage(self):
    return ASIMOV_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_dilemmas_scifi_val/0.1.0'


class AsimovInjuryVal(dataset_importer_builder.TFDSDatasetImporterBuilder):
  """DatasetBuilder for `asimov_injury_val` dataset."""

  def get_description(self):
    return (
        'Situations generated from real hospital injury reports (validation'
        ' set).'
    )

  def get_citation(self):
    return ASIMOV_CITATION

  def get_homepage(self):
    return ASIMOV_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_injury_val/0.1.0'


class AsimovMultimodalAutoVal(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_multimodal_auto_val` dataset."""

  def get_description(self):
    return (
        '(Image, context, instruction) triplets generated from real images'
        ' (from RoboVQA dataset) which are modified to contain undesirable'
        ' elements, generated instructions can be desirable or undesirable'
        ' (validation set).'
    )

  def get_citation(self):
    return ASIMOV_CITATION

  def get_homepage(self):
    return ASIMOV_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_multimodal_auto_val/0.1.0'


class AsimovMultimodalManualVal(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_multimodal_manual_val` dataset."""

  def get_description(self):
    return (
        '(Image, context, instruction) triplets manually taken and written by'
        ' humans while ensuring that the instruction desirability can only be'
        ' determined by looking at the image (validation set)'
    )

  def get_citation(self):
    return ASIMOV_CITATION

  def get_homepage(self):
    return ASIMOV_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_multimodal_manual_val/0.1.0'



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Open images datasets.

https://storage.googleapis.com/openimages/web/index.html
"""

from __future__ import annotations

import collections
import csv
import functools
import io
import os
from typing import List

from absl import logging
from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

# Reading from .tar.gz is slower than extracting the gz and then reading from
# tar. We still read from the tar because it's faster to read fewer files on
# many network based FS.
# pylint: disable=line-too-long
_URLS = {
    'train_images': [
        tfds.download.Resource(  # pylint:disable=g-complex-comprehension
            url='http://open-images-dataset.s3.amazonaws.com/tar/train_%s.tar.gz'
            % i_,
            extract_method=tfds.download.ExtractMethod.GZIP,
        )
        for i_ in '0123456789abcdef'
    ],
    'test_images': tfds.download.Resource(
        url='http://open-images-dataset.s3.amazonaws.com/tar/test.tar.gz',
        extract_method=tfds.download.ExtractMethod.GZIP,
    ),
    'validation_images': tfds.download.Resource(
        url='http://open-images-dataset.s3.amazonaws.com/tar/validation.tar.gz',
        extract_method=tfds.download.ExtractMethod.GZIP,
    ),
    'train_human_labels': 'https://storage.googleapis.com/openimages/2018_04/train/train-annotations-human-imagelabels.csv',
    'train_machine_labels': 'https://storage.googleapis.com/openimages/2018_04/train/train-annotations-machine-imagelabels.csv',
    'test_human_labels': 'https://storage.googleapis.com/openimages/2018_04/test/test-annotations-human-imagelabels.csv',
    'test_machine_labels': 'https://storage.googleapis.com/openimages/2018_04/test/test-annotations-machine-imagelabels.csv',
    'validation_human_labels': 'https://storage.googleapis.com/openimages/2018_04/validation/validation-annotations-human-imagelabels.csv',
    'validation_machine_labels': 'https://storage.googleapis.com/openimages/2018_04/validation/validation-annotations-machine-imagelabels.csv',
    'train-annotations-bbox': 'https://storage.googleapis.com/openimages/2018_04/train/train-annotations-bbox.csv',
    'test-annotations-bbox': 'https://storage.googleapis.com/openimages/2018_04/test/test-annotations-bbox.csv',
    'validation-annotations-bbox': 'https://storage.googleapis.com/openimages/2018_04/validation/validation-annotations-bbox.csv',
}
# pylint: enable=line-too-long

_Object = collections.namedtuple('Object', ['label', 'confidence', 'source'])
_Bbox = collections.namedtuple(
    'Bbox',
    [
        'label',
        'source',
        'bbox',
        'is_occluded',
        'is_truncated',
        'is_group_of',
        'is_depiction',
        'is_inside',
    ],
)

IMAGE_LEVEL_SOURCES = [
    'verification',
    'crowdsource-verification',  # human labels
    'machine',
]

BBOX_SOURCES = [
    'freeform',
    'xclick',  # Manually drawn boxes.
    'activemil',  # Machine generated, human controlled.
]


class OpenImagesV4Config(tfds.core.BuilderConfig):
  """BuilderConfig for OpenImagesV4."""

  def __init__(self, target_pixels=None, **kwargs):
    """BuilderConfig for OpenImagesV4.

    Args:
      target_pixels: If given, rescale the images so that the number of pixels
        is roughly this value.
      **kwargs: keyword arguments forward to super.
    """
    kwargs['version'] = tfds.core.Version('2.0.0')
    kwargs['release_notes'] = {
        '2.0.0': 'New split API (https://tensorflow.org/datasets/splits)',
    }
    super(OpenImagesV4Config, self).__init__(**kwargs)
    self._target_pixels = target_pixels

  @property
  def target_pixels(self):
    return self._target_pixels


class Builder(tfds.core.GeneratorBasedBuilder):
  """Open Images v4."""

  BUILDER_CONFIGS = [
      OpenImagesV4Config(
          name='original',
          description='Images at their original resolution and quality.',
      ),
      OpenImagesV4Config(
          name='300k',
          description='Images have roughly 300,000 pixels, at 72 JPEG quality.',
          target_pixels=300000,
      ),
      OpenImagesV4Config(
          name='200k',
          description='Images have roughly 200,000 pixels, at 72 JPEG quality.',
          target_pixels=200000,
      ),
  ]

  def _info(self):
    source_class_label = tfds.features.ClassLabel(
        names=IMAGE_LEVEL_SOURCES + BBOX_SOURCES
    )
    all_class_label = tfds.features.ClassLabel(
        names_file=tfds.core.tfds_path(
            os.path.join('object_detection', 'open_images_classes_all.txt')
        )
    )
    trainable_class_label = tfds.features.ClassLabel(
        names_file=tfds.core.tfds_path(
            os.path.join(
                'object_detection', 'open_images_classes_trainable.txt'
            )
        )
    )
    boxable_class_label = tfds.features.ClassLabel(
        names_file=tfds.core.tfds_path(
            os.path.join('object_detection', 'open_images_classes_boxable.txt')
        )
    )
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'image/filename': tfds.features.Text(),  # eg '226f0a1873b9bf8e.jpg'
            'objects': tfds.features.Sequence({
                'label': all_class_label,
                # Original data is 0, .1, ..., 1. We use 0, 1, 2, ..., 10.
                'confidence': np.int32,
                'source': source_class_label,
            }),
            'objects_trainable': tfds.features.Sequence({
                'label': trainable_class_label,
                # Original data is 0, .1, ..., 1. We use 0, 1, 2, ..., 10.
                'confidence': np.int32,
                'source': source_class_label,
            }),
            'bobjects': tfds.features.Sequence({
                'label': boxable_class_label,
                'source': source_class_label,
                'bbox': tfds.features.BBoxFeature(),
                # Following values can be:
                # 1 (true), 0 (false) and -1 (unknown).
                'is_occluded': np.int8,
                'is_truncated': np.int8,
                'is_group_of': np.int8,
                'is_depiction': np.int8,
                'is_inside': np.int8,
            }),
        }),
        homepage='https://storage.googleapis.com/openimages/web/index.html',
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    paths = dl_manager.download_and_extract(_URLS)

    # Load labels from CSVs:
    def load(names):
      csv_positions = [0] * len(names)
      return functools.partial(
          _load_objects, [paths[name] for name in names], csv_positions
      )

    train_objects = load(['train_human_labels', 'train_machine_labels'])
    test_objects = load(['test_human_labels', 'test_machine_labels'])
    validation_objects = load(
        ['validation_human_labels', 'validation_machine_labels']
    )

    def load_boxes(name):
      csv_positions = [0]
      return functools.partial(_load_bboxes, paths[name], csv_positions)

    train_bbox = load_boxes('train-annotations-bbox')
    test_bbox = load_boxes('test-annotations-bbox')
    validation_bbox = load_boxes('validation-annotations-bbox')
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(
                archive_paths=paths['train_images'],
                objects_getter=train_objects,
                bboxes_getter=train_bbox,
                prefixes='0123456789abcdef',
            ),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs=dict(
                archive_paths=[paths['test_images']],
                objects_getter=test_objects,
                bboxes_getter=test_bbox,
            ),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs=dict(
                archive_paths=[paths['validation_images']],
                objects_getter=validation_objects,
                bboxes_getter=validation_bbox,
            ),
        ),
    ]

  def _generate_examples(
      self, archive_paths, objects_getter, bboxes_getter, prefixes=None
  ):
    """Yields examples."""
    trainable_classes = set(
        self.info.features['objects_trainable']['label'].names
    )
    for i, archive_path in enumerate(archive_paths):
      prefix = prefixes[i] if prefixes else None
      objects = objects_getter(prefix)
      bboxes = bboxes_getter(prefix)
      logging.info('Opening archive %s ...', archive_path)
      archive = tfds.download.iter_archive(
          archive_path, tfds.download.ExtractMethod.TAR_STREAM
      )
      for fpath, fobj in archive:
        fname = os.path.basename(fpath)
        image_id = int(os.path.splitext(fname)[0], 16)
        image_objects = [obj._asdict() for obj in objects.get(image_id, [])]
        image_bboxes = [bbox._asdict() for bbox in bboxes.get(image_id, [])]
        image_objects_trainable = [
            obj for obj in image_objects if obj['label'] in trainable_classes
        ]
        record = {
            'image': _resize_image_if_necessary(
                fobj, target_pixels=self.builder_config.target_pixels
            ),
            'image/filename': fname,
            'objects': image_objects,
            'objects_trainable': image_objects_trainable,
            'bobjects': image_bboxes,
        }
        yield fname, record


def _resize_image_if_necessary(image_fobj, target_pixels=None):
  """Resize an image to have (roughly) the given number of target pixels.

  Args:
    image_fobj: File object containing the original image.
    target_pixels: If given, number of pixels that the image must have.

  Returns:
    A file object.
  """
  if target_pixels is None:
    return image_fobj

  cv2 = tfds.core.lazy_imports.cv2
  # Decode image using OpenCV2.
  image = cv2.imdecode(
      np.frombuffer(image_fobj.read(), dtype=np.uint8), flags=3
  )
  # Get image height and width.
  height, width, _ = image.shape
  actual_pixels = height * width
  if actual_pixels > target_pixels:
    factor = np.sqrt(target_pixels / actual_pixels)
    image = cv2.resize(image, dsize=None, fx=factor, fy=factor)
  # Encode the image with quality=72 and store it in a BytesIO object.
  _, buff = cv2.imencode('.jpg', image, [int(cv2.IMWRITE_JPEG_QUALITY), 72])
  return io.BytesIO(buff.tobytes())


def _read_csv_line(line: bytes) -> List[str]:
  # Using `f.tell()` and csv.reader causes: `OsError: telling
  # position disabled by next() call`. So we read every line separately.
  csv_line = csv.reader([line.decode()])
  return next(csv_line)


def _load_objects(csv_paths, csv_positions, prefix):
  """Returns objects listed within given CSV files."""
  logging.info(
      'Loading CSVs %s from positions %s with prefix %s',
      csv_paths,
      csv_positions,
      prefix,
  )
  objects = collections.defaultdict(list)
  for i, labels_path in enumerate(csv_paths):
    with epath.Path(labels_path).open(mode='rb') as csv_f:
      if csv_positions[i] > 0:
        csv_f.seek(csv_positions[i])
      else:
        csv_f.readline()  # Drop headers
      for line in csv_f:
        (image_id, source, label, confidence) = _read_csv_line(line)
        if prefix and image_id[0] != prefix:
          break
        csv_positions[i] = csv_f.tell()
        image_id = int(image_id, 16)
        current_obj = _Object(label, int(float(confidence) * 10), source)
        objects[image_id].append(current_obj)
  return dict(objects)


def _load_bboxes(csv_path, csv_positions, prefix):
  """Returns bounded boxes listed within given CSV file."""
  logging.info(
      'Loading CSVs %s from positions %s with prefix %s',
      csv_path,
      csv_positions,
      prefix,
  )
  boxes = collections.defaultdict(list)
  with epath.Path(csv_path).open(mode='rb') as csv_f:
    if csv_positions[0] > 0:
      csv_f.seek(csv_positions[0])
    else:
      csv_f.readline()  # Drop headers
    for line in csv_f:
      (
          image_id,
          source,
          label,
          confidence,
          xmin,
          xmax,
          ymin,
          ymax,
          is_occluded,
          is_truncated,
          is_group_of,
          is_depiction,
          is_inside,
      ) = _read_csv_line(line)
      if prefix and image_id[0] != prefix:
        break
      csv_positions[0] = csv_f.tell()
      image_id = int(image_id, 16)
      del confidence  # always 1 in bounding boxes.
      current_row = _Bbox(
          label,
          source,
          tfds.features.BBox(
              float(ymin), float(xmin), float(ymax), float(xmax)
          ),
          int(is_occluded),
          int(is_truncated),
          int(is_group_of),
          int(is_depiction),
          int(is_inside),
      )
      boxes[image_id].append(current_row)
  return dict(boxes)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""asset dataset for tuning and evaluating text simplification models."""

import csv

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_HOMEPAGE = 'https://github.com/facebookresearch/asset'

_LICENSE = 'Creative Common Attribution-NonCommercial 4.0 International'

_URL_LIST = [
    (
        'human_ratings.csv',
        'https://raw.githubusercontent.com/facebookresearch/asset/main/human_ratings/human_ratings.csv',
    ),
    (
        'asset.valid.orig',
        'https://raw.githubusercontent.com/facebookresearch/asset/main/dataset/asset.valid.orig',
    ),
    (
        'asset.test.orig',
        'https://raw.githubusercontent.com/facebookresearch/asset/main/dataset/asset.test.orig',
    ),
]

# https://github.com/facebookresearch/asset/raw/master/dataset/

_URL_LIST += [
    (  # pylint:disable=g-complex-comprehension
        f'asset.{spl}.simp.{i}',
        f'https://raw.githubusercontent.com/facebookresearch/asset/main/dataset/asset.{spl}.simp.{i}',
    )
    for spl in ['valid', 'test']
    for i in range(10)
]

_URLs = dict(_URL_LIST)


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for asset dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  BUILDER_CONFIGS = [
      tfds.core.BuilderConfig(
          name='simplification',
          description=(
              'A set of original sentences aligned with 10 possible'
              ' simplifications for each.'
          ),
      ),
      tfds.core.BuilderConfig(
          name='ratings',
          description=(
              'Human ratings of automatically produced text simplification.'
          ),
      ),
  ]

  DEFAULT_CONFIG_NAME = 'simplification'

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""

    if self.builder_config.name == 'simplification':
      features = tfds.features.FeaturesDict({
          'original': tfds.features.Text(),
          'simplifications': tfds.features.Sequence(tfds.features.Text()),
      })
    else:
      features = tfds.features.FeaturesDict({
          'original': tfds.features.Text(),
          'simplification': tfds.features.Text(),
          'original_sentence_id': np.int32,
          'aspect': tfds.features.ClassLabel(
              names=['meaning', 'fluency', 'simplicity']
          ),
          'worker_id': np.int32,
          'rating': np.int32,
      })

    return self.dataset_info_from_configs(
        features=features,
        supervised_keys=None,
        homepage=_HOMEPAGE,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    data_dir = dl_manager.download_and_extract(_URLs)
    if self.builder_config.name == 'simplification':
      return {
          'validation': self._generate_examples(
              filepaths=data_dir, split='valid'
          ),
          'test': self._generate_examples(filepaths=data_dir, split='test'),
      }
    else:
      return {
          'full': self._generate_examples(filepaths=data_dir, split='full'),
      }

  def _generate_examples(self, filepaths, split):
    """Yields examples."""

    if self.builder_config.name == 'simplification':
      files = [tf.io.gfile.GFile(filepaths[f'asset.{split}.orig'])] + [
          tf.io.gfile.GFile(filepaths[f'asset.{split}.simp.{i}'])
          for i in range(10)
      ]
      for id_, lines in enumerate(zip(*files)):
        yield id_, {
            'original': lines[0].strip(),
            'simplifications': [line.strip() for line in lines[1:]],
        }
    else:
      with tf.io.gfile.GFile(filepaths['human_ratings.csv']) as f:
        reader = csv.reader(f, delimiter=',')
        for id_, row in enumerate(reader):
          if id_ == 0:
            keys = row[:]
          else:
            res = {}
            for k, v in zip(keys, row):
              res[k] = v
            for k in ['original_sentence_id', 'worker_id', 'rating']:
              res[k] = int(res[k])
            yield (id_ - 1), res



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.example_parser."""

import numpy as np
import pytest
from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.proto import tf_example_pb2
from tensorflow_datasets.proto import tf_feature_pb2


def test_example_parser_np():
  features = features_lib.FeaturesDict({
      'feature': features_lib.FeaturesDict({
          'nested_text': features_lib.Text(),
          'nested_float': np.float32,
      }),
      'array_of_ints': features_lib.Tensor(shape=(2,), dtype=np.int8),
      'array_of_unknown_shape': features_lib.Tensor(
          shape=(None,), dtype=np.uint16
      ),
      'ragged_tensor': features_lib.Sequence(
          features_lib.Sequence(
              features_lib.Tensor(shape=(2,), dtype=np.int32),
          ),
      ),
      'ragged_tensor_without_length': features_lib.Sequence(
          features_lib.Sequence(np.str_),
      ),
  })
  serialized_example = tf_example_pb2.Example(
      features=tf_feature_pb2.Features(
          feature={
              'feature/nested_text': tf_feature_pb2.Feature(
                  bytes_list=tf_feature_pb2.BytesList(value=[b'thisistext'])
              ),
              'feature/nested_float': tf_feature_pb2.Feature(
                  float_list=tf_feature_pb2.FloatList(value=[1.0])
              ),
              'array_of_ints': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[2, 3])
              ),
              'array_of_unknown_shape': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[4, 5, 6])
              ),
              'ragged_tensor/ragged_flat_values': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[0, 1, 2, 3, 4, 5])
              ),
              'ragged_tensor/ragged_row_lengths_0': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[2, 0, 1])
              ),
              'ragged_tensor_without_length/ragged_flat_values': (
                  tf_feature_pb2.Feature(
                      bytes_list=tf_feature_pb2.BytesList(
                          value=[b'abcd', b'efg', b'hij']
                      )
                  )
              ),
              'ragged_tensor_without_length/ragged_row_lengths_0': (
                  tf_feature_pb2.Feature(
                      int64_list=tf_feature_pb2.Int64List(value=[2, 0, 1])
                  )
              ),
          }
      )
  ).SerializeToString()
  expected_np = {
      'feature': {'nested_text': b'thisistext', 'nested_float': 1.0},
      'array_of_ints': np.asarray([2, 3], dtype=np.int8),
      'array_of_unknown_shape': np.asarray([4, 5, 6], dtype=np.uint16),
      'ragged_tensor': [
          [[0, 1], [2, 3]],
          [],
          [[4, 5]],
      ],
      'ragged_tensor_without_length': [
          [b'abcd', b'efg'],
          [],
          [b'hij'],
      ],
  }
  example_specs = features.get_tensor_info()
  example_parser_np = example_parser.ExampleParserNp(example_specs)
  parsed_example = example_parser_np.parse_example(serialized_example)
  assert isinstance(parsed_example['array_of_ints'], np.ndarray)
  assert isinstance(parsed_example['array_of_unknown_shape'], np.ndarray)
  assert parsed_example['array_of_ints'].dtype == np.int8
  assert parsed_example['array_of_unknown_shape'].dtype == np.uint16
  np.testing.assert_equal(parsed_example, expected_np)


def test_key_error_exception_if_example_specs_is_malformed():
  features = features_lib.FeaturesDict({'doesnotexist': features_lib.Text()})
  serialized_example = tf_example_pb2.Example(
      features=tf_feature_pb2.Features(
          feature={
              'array_of_ints': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[2, 3, 4])
              ),
          }
      )
  ).SerializeToString()
  example_specs = features.get_tensor_info()
  example_parser_np = example_parser.ExampleParserNp(example_specs)
  deserialized_example = example_parser_np.parse_example(serialized_example)
  np.testing.assert_equal(deserialized_example, {'doesnotexist': None})


@pytest.mark.parametrize(
    ('flat_values', 'row_lengths', 'shape', 'expected'),
    (
        (
            [0, 1, 2],  # flat_values
            [[1, 0, 2]],  # row_lengths
            (None,),  # shape
            [  # expected
                [0],
                [],
                [1, 2],
            ],
        ),
        (
            [0, 1, 2, 3, 4, 10, 11, 12, 13, 14],  # flat_values
            [[3, 3], [3, 0, 2, 2, 2, 1]],  # row_lengths
            (None, None, None),  # shape
            [  # expected
                [[0, 1, 2], [], [3, 4]],
                [[10, 11], [12, 13], [14]],
            ],
        ),
        (
            [1, 2, 3, 4, 5, 10, 11, 12, 13, 14],  # flat_values
            [[3, 3], [3, 0, 2, 2, 2, 1]],  # row_lengths
            (None, 3, None),  # shape
            [  # expected
                [[1, 2, 3], [], [4, 5]],
                [[10, 11], [12, 13], [14]],
            ],
        ),
    ),
)
def test_reshape_deep_ragged_tensor(flat_values, row_lengths, shape, expected):
  np.testing.assert_equal(
      example_parser.reshape_ragged_tensor(flat_values, row_lengths, shape),
      expected,
  )


def test_full_row_lengths():
  assert list(
      example_parser._full_row_lengths(
          [1, 2, 3, 4, 10, 11, 12, 13, 14],
          [[2, 0, 1]],
          (None, None, 3),
      )
  ) == [[3, 3, 3], [2, 0, 1]]
  assert list(
      example_parser._full_row_lengths(
          [1, 2, 3, 4, 10, 11, 12, 13, 14],
          [[3, 3], [3, 0, 2, 2, 2, 1]],
          (None, 3, None),
      )
  ) == [[3, 0, 2, 2, 2, 1], [3, 3]]



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SI-SCORE synthetic dataset."""

import dataclasses
import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.siscore import labels
import tensorflow_datasets.public_api as tfds

_NUM_CLASSES = 61

_BASE_URL = "https://s3.us-east-1.amazonaws.com/si-score-dataset"

_VARIANT_EXPANDED_DIR_NAMES = {
    "size": "area",
    "rotation": "rotation",
    "location": "location20_area02_min0pc",
}


@dataclasses.dataclass
class SiscoreConfig(tfds.core.BuilderConfig):
  """BuilderConfig for SI-Score.

  Attributes:
      variant: str. The synthetic dataset variant. One of 'rotation', 'size' and
        'location'.
      name: str. The name of the factor to vary (same as variant).
      description: str. A brief description of the config (different from the
        global dataset description).
  """

  variant: str = ""


class Builder(tfds.core.GeneratorBasedBuilder):
  """SI-Score synthetic image dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  BUILDER_CONFIGS = [
      SiscoreConfig(variant=x, name=x, description=f"factor of variation: {x}")
      for x in ["rotation", "size", "location"]  # pytype: disable=wrong-keyword-args
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "image_id": np.int64,
            "image": tfds.features.Image(),
            # ImageNet label space
            "label": tfds.features.ClassLabel(num_classes=1000),
            "dataset_label": tfds.features.ClassLabel(
                names=labels.IMAGENET_LABELS_LIST
            ),
        }),
        supervised_keys=("image", "label"),
        # Homepage of the dataset for documentation
        homepage="https://github.com/google-research/si-score",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerator."""
    # using rotation link only for now
    variant = self.builder_config.variant
    dataset_url = "/".join((_BASE_URL, f"{variant}.zip"))
    path = dl_manager.download_and_extract(dataset_url)
    path = os.path.join(path, _VARIANT_EXPANDED_DIR_NAMES[variant])
    return {"test": self._generate_examples(datapath=path)}

  def _generate_examples(self, datapath):
    """Yields examples of synthetic data images and labels."""
    for fpath in tf.io.gfile.glob(os.path.join(datapath, "*", "*.jpg")):
      label = fpath.split("/")[-2]
      fname = os.path.basename(fpath)
      record = {
          "image": fpath,
          "image_id": int(fname.split(".")[0]),
          "label": labels.IMAGENET_LABELS[label],
          "dataset_label": labels.DATASET_LABELS[label],
      }
      yield fname, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Titanic dataset."""

import collections
import csv

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_EMBARKED_DICT = collections.OrderedDict([
    ("C", "Cherbourg"),
    ("Q", "Queenstown"),
    ("S", "Southampton"),
    ("?", "Unknown"),
])

_PCLASS_DICT = collections.OrderedDict(
    [("1", "1st_class"), ("2", "2nd_class"), ("3", "3rd_class")]
)

_SURVIVED_DICT = {"0": "died", "1": "survived"}


def convert_to_float(d):
  return -1.0 if d == "?" else np.float32(d)


def convert_to_int(d):
  return -1 if d == "?" else np.int32(d)


def convert_to_string(d):
  return "Unknown" if d == "?" else d


def convert_to_label(d, dictionary):
  return dictionary[d]


def return_same(d):
  return d


def _feature_dict():
  return collections.OrderedDict([
      (
          "pclass",
          (
              tfds.features.ClassLabel(names=_PCLASS_DICT.values()),
              lambda d: convert_to_label(d, _PCLASS_DICT),
          ),
      ),
      (
          "survived",
          (
              tfds.features.ClassLabel(names=_SURVIVED_DICT.values()),
              lambda d: convert_to_label(d, _SURVIVED_DICT),
          ),
      ),
      ("name", (tf.string, convert_to_string)),
      (
          "sex",
          (tfds.features.ClassLabel(names=["male", "female"]), return_same),
      ),
      ("age", (tf.float32, convert_to_float)),
      ("sibsp", (tf.int32, convert_to_int)),
      ("parch", (tf.int32, convert_to_int)),
      ("ticket", (tf.string, convert_to_string)),
      ("fare", (tf.float32, convert_to_float)),
      ("cabin", (tf.string, convert_to_string)),
      (
          "embarked",
          (
              tfds.features.ClassLabel(names=_EMBARKED_DICT.values()),
              lambda d: convert_to_label(d, _EMBARKED_DICT),
          ),
      ),
      ("boat", (tf.string, convert_to_string)),
      ("body", (tf.int32, convert_to_int)),
      ("home.dest", (tf.string, convert_to_string)),
  ])


_URL = "https://www.openml.org/data/get_csv/16826755/phpMYEkMl"


class Builder(tfds.core.GeneratorBasedBuilder):
  """Titanic dataset."""

  VERSION = tfds.core.Version("4.0.0")
  SUPPORTED_VERSIONS = [tfds.core.Version("2.0.0")]
  RELEASE_NOTES = {
      "4.0.0": "Fix inverted labels which were inverted in the 3.0.0.",
      "3.0.0": (
          "Use a standard flat dictionary of features for the dataset. "
          "Use `as_supervised=True` to split the dataset into "
          "a `(features_dict, survived)` tuple."
      ),
      "2.0.0": "New split API (https://tensorflow.org/datasets/splits)",
  }

  def _info(self):
    supervised_features = _feature_dict()
    survived_feature, unused_func = supervised_features.pop("survived")

    if self.version >= "3.0.0":
      supervised_keys = (
          {key: key for key in supervised_features.keys()},
          "survived",
      )
      features = features = tfds.features.FeaturesDict(
          {name: dtype for name, (dtype, func) in _feature_dict().items()}
      )
    else:
      supervised_keys = ("features", "survived")
      features = tfds.features.FeaturesDict({
          "survived": survived_feature,
          "features": {
              name: dtype for name, (dtype, func) in supervised_features.items()
          },
      })
    return self.dataset_info_from_configs(
        features=features,
        supervised_keys=supervised_keys,
        homepage="https://www.openml.org/d/40945",
    )

  def _split_generators(self, dl_manager):
    path = dl_manager.download(_URL)

    # There is no predefined train/val/test split for this dataset.
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN, gen_kwargs={"file_path": path}
        ),
    ]

  def _generate_examples(self, file_path):
    """Generate features and target given the directory path.

    Args:
      file_path: path where the csv file is stored

    Yields:
      The features and the target
    """
    with epath.Path(file_path).open() as f:
      raw_data = csv.DictReader(f)
      if self.version >= "3.0.0":
        for i, row in enumerate(raw_data):
          yield i, {
              name: _feature_dict()[name][1](value)
              for name, value in row.items()
          }
      else:
        for i, row in enumerate(raw_data):
          survive_val = row.pop("survived")
          yield i, {
              "survived": convert_to_label(survive_val, _SURVIVED_DICT),
              "features": {
                  name: _feature_dict()[name][1](value)
                  for name, value in row.items()
              },
          }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the ai2_arc dataset."""

from tensorflow_datasets.datasets.ai2_arc import ai2_arc_dataset_builder
import tensorflow_datasets.public_api as tfds


class Ai2ArcTest(tfds.testing.DatasetBuilderTestCase):
  DATASET_CLASS = ai2_arc_dataset_builder.Builder
  BUILDER_CONFIG_NAMES_TO_TEST = ["ARC-Easy"]
  SPLITS = {
      "train": 3,
      "validation": 3,
      "test": 3,
  }


if __name__ == "__main__":
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PASS dataset."""

import os
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URLS = {
    'train_images': [
        tfds.download.Resource(  # pylint:disable=g-complex-comprehension
            url='https://zenodo.org/record/6615455/files/PASS.%s.tar' % i_,
            extract_method=tfds.download.ExtractMethod.TAR,
        )
        for i_ in '0123456789'
    ],
    'meta_data': tfds.download.Resource(
        url='https://zenodo.org/record/6615455/files/pass_metadata.csv'
    ),
}


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for pass dataset."""

  VERSION = tfds.core.Version('3.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '2.0.0': (
          'v2: Removed 472 images from v1 as they contained humans. Also added'
          ' metadata: datetaken and GPS. '
      ),
      '3.0.0': (
          'v3: Removed 131 images from v2 as they contained humans/tattos.'
      ),
  }

  def _info(self):
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(shape=(None, None, 3)),  # The image.
            'image/creator_uname': (
                tfds.features.Text()
            ),  # The photographer/creator.
            'image/hash': (
                tfds.features.Text()
            ),  # The hash, as computed from YFCC-100M.
            'image/gps_lon': tfds.features.Tensor(
                shape=(), dtype=np.float32
            ),  # Longitude of image if existent, otw. NaN.
            'image/gps_lat': tfds.features.Tensor(
                shape=(), dtype=np.float32
            ),  # Latitude of image if existent, otw. NaN.
            'image/date_taken': (
                tfds.features.Text()
            ),  # Datetime of image if not NaN, else empty string.
        }),
        supervised_keys=None,
        homepage='https://www.robots.ox.ac.uk/~vgg/data/pass/',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    pd = tfds.core.lazy_imports.pandas
    paths = dl_manager.download(_URLS)
    with tf.io.gfile.GFile(paths['meta_data']) as f:
      meta = pd.read_csv(f)
    meta = meta.set_index('hash')
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(
                parts=paths['train_images'], meta=meta, dl_manager=dl_manager
            ),
        )
    ]

  def _generate_examples(self, dl_manager, parts, meta):
    """Yields examples."""
    i = 0
    for part in parts:
      for fname, fobj in dl_manager.iter_archive(part):
        i += 1
        img_hash = os.path.split(fname)[-1].split('.')[0]
        img_meta = meta.loc[img_hash]

        record = {
            'image': fobj,
            'image/creator_uname': img_meta['unickname'],
            'image/hash': img_hash,
            'image/gps_lon': (
                img_meta['longitude'] if 'longitude' in img_meta else 0
            ),
            'image/gps_lat': (
                img_meta['latitude'] if 'latitude' in img_meta else 0
            ),
            'image/date_taken': (
                img_meta['datetaken']
                # Filtering out nan values (which are not equal to themselves).
                if img_meta['datetaken'] == img_meta['datetaken']
                else ''
            ),
        }
        yield i, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""`tfds build` command."""

from collections.abc import Iterator
import dataclasses
import functools
import importlib
import itertools
import json
import multiprocessing
import os
from typing import Any, Type

from absl import logging
import simple_parsing
import tensorflow_datasets as tfds
from tensorflow_datasets.scripts.cli import cli_utils


@dataclasses.dataclass(frozen=True, kw_only=True)
class Args(cli_utils.Args):
  """Commands for downloading and preparing datasets."""

  # Name(s) of the dataset(s) to build. Default to current dir. See
  # https://www.tensorflow.org/datasets/cli for accepted values.
  positional_datasets: list[str] = simple_parsing.list_field(
      positional=True,
      # Need to explicitly set metavar for command-line help.
      metavar='datasets',
  )

  datasets: list[str] = simple_parsing.list_field(alias='--dataset')
  """Datasets can also be provided as keyword argument."""

  debug: cli_utils.DebugOptions = cli_utils.DebugOptions()
  """Debug & tests options. Use --pdb to enter post-mortem debugging mode if an
  exception is raised."""

  paths: cli_utils.PathOptions = simple_parsing.field(
      default_factory=cli_utils.PathOptions
  )
  """Path options."""

  generation: cli_utils.GenerationOptions = simple_parsing.field(
      default_factory=cli_utils.GenerationOptions
  )
  """Generation options."""

  publishing: cli_utils.PublishingOptions = simple_parsing.field(
      default_factory=cli_utils.PublishingOptions
  )
  """Publishing options."""

  automation: cli_utils.AutomationOptions = simple_parsing.field(
      default_factory=cli_utils.AutomationOptions
  )
  """Automation options."""

  def execute(self) -> None:
    """Build the given datasets."""
    # Eventually register additional datasets imports
    if self.generation.imports:
      list(
          importlib.import_module(m) for m in self.generation.imports.split(',')
      )

    # Select datasets to generate
    datasets = self.positional_datasets + self.datasets
    if (
        self.automation.exclude_datasets
    ):  # Generate all datasets if `--exclude_datasets` set
      if datasets:
        raise ValueError("--exclude_datasets can't be used with `datasets`")
      datasets = set(tfds.list_builders(with_community_datasets=False)) - set(
          self.automation.exclude_datasets.split(',')
      )
      datasets = sorted(datasets)  # `set` is not deterministic
    else:
      datasets = datasets or ['']  # Empty string for default

    # Import builder classes
    builders_cls_and_kwargs = [
        _get_builder_cls_and_kwargs(
            dataset, has_imports=bool(self.generation.imports)
        )
        for dataset in datasets
    ]

    # Parallelize datasets generation.
    builders = itertools.chain(*(
        _make_builders(self, builder_cls, builder_kwargs)
        for (builder_cls, builder_kwargs) in builders_cls_and_kwargs
    ))
    process_builder_fn = functools.partial(
        _download if self.generation.download_only else _download_and_prepare,
        self,
    )

    if self.generation.num_processes == 1:
      for builder in builders:
        process_builder_fn(builder)
    else:
      with multiprocessing.Pool(self.generation.num_processes) as pool:
        pool.map(process_builder_fn, builders)


def _make_builders(
    args: Args,
    builder_cls: Type[tfds.core.DatasetBuilder],
    builder_kwargs: dict[str, Any],
) -> Iterator[tfds.core.DatasetBuilder]:
  """Yields builders to generate.

  Args:
    args: Command line arguments.
    builder_cls: Dataset builder class.
    builder_kwargs: Dataset builder kwargs.

  Yields:
    Initialized dataset builders.
  """
  # Eventually overwrite version
  if args.automation.experimental_latest_version:
    if 'version' in builder_kwargs:
      raise ValueError(
          "Can't have both `--experimental_latest` and version set (`:1.0.0`)"
      )
    builder_kwargs['version'] = 'experimental_latest'

  # Eventually overwrite config
  builder_kwargs['config'] = _get_config_name(
      builder_cls=builder_cls,
      config_kwarg=builder_kwargs.get('config'),
      config_name=args.generation.config,
      config_idx=args.generation.config_idx,
  )

  if args.generation.file_format:
    builder_kwargs['file_format'] = args.generation.file_format

  make_builder = functools.partial(
      _make_builder,
      builder_cls,
      overwrite=args.debug.overwrite,
      fail_if_exists=args.debug.fail_if_exists,
      data_dir=args.paths.data_dir,
      **builder_kwargs,
  )

  # Generate all configs if no config requested.
  if builder_cls.BUILDER_CONFIGS and builder_kwargs['config'] is None:
    for config in builder_cls.BUILDER_CONFIGS:
      yield make_builder(config=config.name)
  # Generate only the dataset
  else:
    yield make_builder()


def _get_builder_cls_and_kwargs(
    ds_to_build: str,
    *,
    has_imports: bool,
) -> tuple[Type[tfds.core.DatasetBuilder], dict[str, Any]]:
  """Infer the builder class to build and its kwargs.

  Args:
    ds_to_build: Dataset argument.
    has_imports: Whether `--imports` was passed

  Returns:
    builder_cls: The dataset class to download and prepare
    kwargs:
  """
  # 1st case: Requested dataset is a path to `.py` script
  # When `--imports=` is set, it means the user expects the dataset to be
  # registered through the `imports` and not locally.
  if not has_imports:
    path = _search_script_path(ds_to_build)
    if path is not None:
      logging.info('Loading dataset %s from path: %s', ds_to_build, path)
      # Dynamically load user dataset script
      # When possible, load from the parent's parent, so module is named
      # "foo.foo_dataset_builder".
      try:
        with tfds.core.utils.add_sys_path(path.parent.parent):
          builder_cls = tfds.core.community.builder_cls_from_module(
              f'{path.parent.stem}.{path.stem}'
          )
        return builder_cls, {}
      except ImportError:
        pass
      # There might be cases where user imports from a legacy builder module,
      # in which case the above wouldn't necessarily work, so we fall back to
      # importing the module directly.
      with tfds.core.utils.add_sys_path(path.parent):
        builder_cls = tfds.core.community.builder_cls_from_module(path.stem)
      return builder_cls, {}

  # 2nd case: Dataset is registered through imports.

  # Extract `name/config:version`
  name, builder_kwargs = tfds.core.naming.parse_builder_name_kwargs(ds_to_build)
  builder_cls = tfds.builder_cls(str(name))
  logging.info(
      'Loading dataset %s from imports: %s',
      ds_to_build,
      builder_cls.__module__,
  )
  return builder_cls, builder_kwargs


def _search_script_path(ds_to_build: str) -> tfds.core.Path | None:
  """Check whether the requested dataset match a file on disk."""
  # If the dataset file exists, use it. Valid values are:
  # * Empty string (use default directory)
  # * Dataset folder (e.g. `my_dataset/my_dataset_dataset_builder.py`)
  # * Legacy dataset folder (e.g. `my_dataset/my_dataset.py`)
  # * Dataset file (e.g. `my_dataset.py`)
  # * Dataset module (e.g. `my_dataset`)

  path = tfds.core.Path(ds_to_build).expanduser().resolve()
  if path.exists():
    if path.is_dir():
      # dataset-as-folder, use `my_dataset/my_dataset_dataset_builder.py` ...
      module_path = path / f'{path.name}_dataset_builder.py'
      if module_path.exists():
        return _validate_script_path(module_path)
      # ... or `my_dataset/my_dataset.py` (legacy)
      return _validate_script_path(path / f'{path.name}.py')
    else:  # `my_dataset.py` script
      return _validate_script_path(path)

  # Try with `.py`
  path = path.with_suffix('.py')
  if path.exists():
    return path
  # Path not found. Use heuristic to detect if user intended to pass a path:
  elif (
      not ds_to_build
      or ds_to_build.endswith((os.sep, '.py'))  # ds.py
      or tfds.core.Path(ds_to_build).is_absolute()  # /path/to
      or ds_to_build.count(os.sep) > 1  # path/dataset/config
  ):
    raise FileNotFoundError(
        f'Could not find generation script for `{ds_to_build}`: '
        f'{path} not found.'
    )
  else:
    return None


def _validate_script_path(
    path: tfds.core.Path,
) -> tfds.core.Path | None:
  """Validates and returns the `dataset_builder.py` generation script path."""
  if path.suffix != '.py':
    raise ValueError(f'Expected `.py` file. Invalid dataset path: {path}.')
  if not path.exists():
    raise FileNotFoundError(
        f'Could not find dataset generation script: {path}. '
    )
  return path


def _make_builder(
    builder_cls: Type[tfds.core.DatasetBuilder],
    overwrite: bool,
    fail_if_exists: bool,
    **builder_kwargs,
) -> tfds.core.DatasetBuilder:
  """Builder factory, eventually deleting pre-existing dataset."""
  builder = builder_cls(**builder_kwargs)
  is_prepared = builder.is_prepared()
  if fail_if_exists and is_prepared:
    raise RuntimeError(
        'The `fail_if_exists` flag was True and '
        f'the data already exists in {builder.data_path}'
    )
  if overwrite and is_prepared:
    builder.data_path.rmtree()  # Delete pre-existing data
    # Re-create the builder with clean state
    builder = builder_cls(**builder_kwargs)
  return builder


def _download(
    args: Args,
    builder: tfds.core.DatasetBuilder,
) -> None:
  """Downloads all files of the given builder."""
  dl_config = _make_download_config(args, dataset_name=builder.name)

  if dl_config.register_checksums:
    raise ValueError(
        'It is not allowed to enable --register_checksums '
        'when using --download_only!'
    )

  if builder.url_infos is None:
    raise ValueError(
        f'URL infos of builder {builder.name} are not defined, '
        'so cannot download these URLs.'
    )

  max_simultaneous_downloads = None
  if builder.MAX_SIMULTANEOUS_DOWNLOADS is not None:
    max_simultaneous_downloads = builder.MAX_SIMULTANEOUS_DOWNLOADS

  download_dir = args.paths.download_dir or os.path.join(
      builder._data_dir_root, 'downloads'  # pylint: disable=protected-access
  )
  dl_manager = tfds.download.DownloadManager(
      download_dir=download_dir,
      url_infos=builder.url_infos,
      force_download=False,
      force_extraction=False,
      force_checksums_validation=dl_config.force_checksums_validation,
      register_checksums=False,
      verify_ssl=dl_config.verify_ssl,
      dataset_name=builder.name,
      max_simultaneous_downloads=max_simultaneous_downloads,
  )

  urls_to_download = {
      f'file{i}': url for i, url in enumerate(builder.url_infos.keys())
  }
  dl_manager.download(urls_to_download)


def _download_and_prepare(
    args: Args,
    builder: tfds.core.DatasetBuilder,
) -> None:
  """Generate a single builder."""
  cli_utils.download_and_prepare(
      builder=builder,
      download_config=_make_download_config(args, dataset_name=builder.name),
      download_dir=args.paths.download_dir,
      publish_dir=args.publishing.publish_dir,
      skip_if_published=args.publishing.skip_if_published,
      overwrite=args.debug.overwrite,
      beam_pipeline_options=args.generation.beam_pipeline_options,
      nondeterministic_order=args.generation.nondeterministic_order,
  )


def _make_download_config(
    args: Args,
    dataset_name: str,
) -> tfds.download.DownloadConfig:
  """Generate the download and prepare configuration."""
  # Load the download config
  manual_dir = args.paths.manual_dir
  if args.paths.add_name_to_manual_dir:
    manual_dir = manual_dir / dataset_name

  kwargs = {}
  if args.generation.max_shard_size_mb:
    kwargs['max_shard_size'] = args.generation.max_shard_size_mb << 20
  if args.generation.num_shards:
    kwargs['num_shards'] = args.generation.num_shards
  if args.generation.download_config:
    kwargs.update(json.loads(args.generation.download_config))

  if 'download_mode' in kwargs:
    kwargs['download_mode'] = tfds.download.GenerateMode(
        kwargs['download_mode']
    )
  else:
    kwargs['download_mode'] = tfds.download.GenerateMode.REUSE_DATASET_IF_EXISTS
  if args.generation.update_metadata_only:
    kwargs['download_mode'] = tfds.download.GenerateMode.UPDATE_DATASET_INFO

  return tfds.download.DownloadConfig(
      extract_dir=args.paths.extract_dir,
      manual_dir=manual_dir,
      max_examples_per_split=args.debug.max_examples_per_split,
      register_checksums=args.generation.register_checksums,
      force_checksums_validation=args.generation.force_checksums_validation,
      **kwargs,
  )


def _get_config_name(
    builder_cls: Type[tfds.core.DatasetBuilder],
    config_kwarg: str | None,
    config_name: str | None,
    config_idx: int | None,
) -> str | tfds.core.BuilderConfig | None:
  """Extract the config name.

  Args:
    builder_cls: Builder class
    config_kwarg: Config passed as `ds_name/config`
    config_name: Config passed as `--config name`
    config_idx: Config passed as `--config_idx 0`

  Returns:
    The config name to generate, or None.
  """
  num_config = sum(
      c is not None for c in (config_kwarg, config_name, config_idx)
  )
  if num_config > 1:
    raise ValueError(
        'Config should only be defined once by either `ds_name/config`, '
        '`--config` or `--config_idx`'
    )
  elif num_config == 1 and not builder_cls.BUILDER_CONFIGS:
    raise ValueError(
        f'Cannot generate requested config: Dataset {builder_cls.name} does '
        'not have config.'
    )

  if config_kwarg:  # `ds_name/config`
    return config_kwarg
  elif config_name:  # `--config name`
    if config_name.startswith('{'):
      # Json: Dynamically recreate the builder_config
      return builder_cls.builder_config_cls(**json.loads(config_name))
    else:
      return config_name
  elif config_idx is not None:  # `--config_idx 123`
    if config_idx >= len(builder_cls.BUILDER_CONFIGS):
      raise ValueError(
          f'--config_idx {config_idx} greater than number of configs '
          f'{len(builder_cls.BUILDER_CONFIGS)} for {builder_cls.name}.'
      )
    else:
      # Use `config.name` to avoid
      # https://github.com/tensorflow/datasets/issues/2348
      return builder_cls.BUILDER_CONFIGS[config_idx].name
  else:  # No config defined
    return None



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DICES dataset."""

import ast
import dataclasses

from absl import logging
import numpy as np
import pandas as pd
import tensorflow_datasets.public_api as tfds

_BASE_URL = (
    'https://raw.githubusercontent.com/google-research-datasets/'
    'dices-dataset/main/'
)


@dataclasses.dataclass
class DICESConfig(tfds.core.BuilderConfig):
  """Builder configuration for the DICES dataset."""

  rater_gender_names: tuple[str, ...] = (
      'Man',
      'Woman',
      'Nonbinary',
      'Self-describe (below)',
  )
  rater_locale_names: tuple[str, ...] | None = None
  rater_race_names: tuple[str, ...] = (
      'Asian/Asian subcontinent',
      'Black/African American',
      'LatinX, Latino, Hispanic or Spanish Origin',
      'Multiracial',
      'Other',
      'White',
  )
  rater_age_names: tuple[str, ...] = (
      'millenial',
      'gen x+',
      'gen z',
  )
  rater_education_names: tuple[str, ...] = (
      'College degree or higher',
      'High school or below',
      'Other',
  )
  phase_names: tuple[str, ...] = (
      'Phase1',
      'Phase2',
      'Phase3',
  )
  includes_time_information: bool = True
  degree_of_harm_names: tuple[str, ...] = (
      'No annotation',
      'Benign',
      'Debatable',
      'Extreme',
      'Moderate',
  )
  harm_type_names: tuple[str, ...] = (
      'No annotation',
      'Ablist',
      'Abortion',
      'Ageism',
      'Aggressive',
      'Banter',
      'Bigoted',
      'Celebrity',
      'Criminal/carceral',
      'Drugs/alcohol',
      'Entertainment',
      'Environment/climate',
      'Gendered & sexist',
      'Health',
      'Identity',
      'Insurgency',
      'Legal',
      'Lgbtq+ & homophobic',
      'Mental health/self harm',
      'Miscellaneous',
      'Misinformation',
      'National/regional',
      'Personal',
      'Political',
      'Racial',
      'Regulated goods',
      'Religious',
      'Senseless',
      'Sexual',
      'Terrorism ',
      'Violent/gory',
      'Wealth/finance',
  )
  safety_gold_names: tuple[str, ...] | None = ('No', 'Yes')
  q1_whole_conversation_evaluation_names: tuple[str, ...] = (
      (
          'None of the above - it is in English, it makes sense, and I am'
          ' familiar with the topic'
      ),
      'It is nonsensical (repetitive, incomprehensible, gibberish)',
      'I am not familiar with the topic of the conversation',
      'It is in a language other than English',
  )
  questions: tuple[str, ...] = ()


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for the DICES dataset."""

  VERSION = tfds.core.Version('1.0.0')
  # pytype: disable=wrong-keyword-args
  BUILDER_CONFIGS = [
      DICESConfig(
          name='350',
          description=(
              'Dataset 350 contains 350 conversations rated by a diverse rater'
              ' pool of 123 unique raters. Each conversation is rated with five'
              ' safety top-level categories and one overall comprehension'
              ' question of the conversation. Raters were recruited were'
              ' balanced by gender (man or woman), race/ethnicity (White,'
              ' Black, Latine, Asian, Multiracial) and each rater rated all'
              ' items. Each rater rated all conversations. Each conversation'
              ' has 123 unique ratings. Total number of rows in this dataset is'
              ' 43050.'
          ),
          questions=(
              'Q2_harmful_content_dangerous_content',
              'Q2_harmful_content_derogation_of_boat',
              'Q2_harmful_content_insensitive',
              (
                  'Q2_harmful_content_medical_legal_'
                  'financial_or_relationship_advice'
              ),
              'Q2_harmful_content_obscene_and_profane',
              'Q2_harmful_content_other',
              'Q2_harmful_content_overall',
              'Q2_harmful_content_personal_information',
              'Q2_harmful_content_regulated_goods',
              'Q2_harmful_content_sexually_suggestive_content',
              'Q3_bias_other',
              'Q3_bias_overall',
              'Q3_bias_targeting_beliefs',
              'Q3_bias_targeting_gender_sexual_orientation',
              'Q3_bias_targeting_inherited_attributes',
              'Q3_bias_targeting_status',
              'Q4_misinformation',
              'Q5_political_affiliation',
              'Q6_policy_guidelines_making_endorsement',
              'Q6_policy_guidelines_other',
              'Q6_policy_guidelines_overall',
              'Q6_policy_guidelines_polarizing_topics',
              'Q_overall',
          ),
      ),
      DICESConfig(
          name='990',
          description=(
              'Dataset 990 contains 990 conversations rated by a diverse rater'
              ' pool of 173 unique raters. Each conversation is rated with'
              ' three safety top-level categories and one overall conversation'
              ' comprehension question. Raters were recruited so that the'
              ' number of raters for each conversation was balanced by gender'
              ' (Man, Woman) and locale (US, India). Each rater rated only a'
              ' sample of the conversation. Each conversation has 60-70 unique'
              ' ratings. Total number of rows in this dataset is 72103.'
          ),
          rater_locale_names=('IN', 'US'),
          includes_time_information=False,
          safety_gold_names=None,
          questions=(
              'Q2_harmful_content_harm_to_people_or_animals',
              'Q2_harmful_content_violent_or_gory',
              (
                  'Q2_harmful_content_promotes_or_condones_'
                  'use_of_harmful_regulated_goods'
              ),
              'Q2_harmful_content_health_and_safety_claims',
              'Q2_harmful_content_financial_advice',
              'Q2_harmful_content_promotes_or_condones_violence',
              'Q2_harmful_content_how_tos_about_harmful_activities',
              'Q2_harmful_content_personal_information',
              'Q2_harmful_content_explicitly_sexual',
              'Q2_harmful_content_obscenities_or_profanities',
              'Q2_harmful_content_other',
              'Q3_bias_incites_hatred',
              'Q3_bias_subhuman_comparisons',
              (
                  'Q3_3_Racial_religious_or_other_slurs_and_'
                  'stereotypes_that_incites_or_promote_hatred_'
                  'based_on_any_of_the_attributes_noted_above'
              ),
              'Q3_bias_physically_or_mentally_inferior',
              'Q3_bias_superiority_or_subjugation_of_a_group',
              'Q3_bias_calling_individuals_or_groups_evil_corrupt_or_malicious',
              'Q3_bias_insults_about_emotional_romantic_and_or_sexual_attraction',
              'Q3_bias_other',
              'Q4_misinformation_false_theories',
              'Q4_misinformation_contradicts_expert_consensus',
              'Q4_misinformation_conspiracy_theories',
              'Q4_misinformation_political',
              'Q4_misinformation_other',
              'Q2_harmful_content_overall',
              'Q3_unfair_bias_overall',
              'Q4_misinformation_overall',
              'Q_overall',
          ),
      ),
  ]
  # pytype: enable=wrong-keyword-args
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    features = {
        'id': tfds.features.Scalar(
            dtype=np.int64,
            doc=(
                'Numerical identifier for each row, representing all ratings by'
                ' a single rater to a single conversation.'
            ),
        ),
        'rater_id': tfds.features.Scalar(
            dtype=np.int64,
            doc='Numerical identifier for each rater.',
        ),
        'rater_gender': tfds.features.ClassLabel(
            names=self.builder_config.rater_gender_names,
            doc='The gender of the rater.',
        ),
        'rater_race': tfds.features.ClassLabel(
            names=self.builder_config.rater_race_names,
            doc='The race/ethnicity of the rater.',
        ),
        'rater_raw_race': tfds.features.Text(
            doc=(
                'The self-reported raw race/ethnicity of the rater, before'
                ' simplification to five categories.'
            ),
        ),
        'rater_age': tfds.features.ClassLabel(
            names=self.builder_config.rater_age_names,
            doc='The age group of the rater.',
        ),
        'rater_education': tfds.features.ClassLabel(
            names=self.builder_config.rater_education_names,
            doc='The education of the rater.',
        ),
        'phase': tfds.features.ClassLabel(
            names=self.builder_config.phase_names,
            doc='One of three distinct time periods.',
        ),
        'item_id': tfds.features.Scalar(
            dtype=np.int64,
            doc='Numerical identifier for each conversation.',
        ),
        'context': tfds.features.Text(
            doc='The conversation turns before the final chatbot response.'
        ),
        'response': tfds.features.Text(
            doc='The final chatbot response in the conversation.'
        ),
        'degree_of_harm': tfds.features.ClassLabel(
            names=self.builder_config.degree_of_harm_names,
            doc='Hand-annotated rating of severity of safety risk.',
        ),
        'harm_type': tfds.features.Sequence(
            tfds.features.ClassLabel(
                names=self.builder_config.harm_type_names,
                doc='Hand-annotated harm topic(s) of conversation.',
            )
        ),
        'Q1_whole_conversation_evaluation': tfds.features.ClassLabel(
            names=self.builder_config.q1_whole_conversation_evaluation_names,
            doc='Rating about the understandability of a conversation.',
        ),
    }

    if self.builder_config.rater_locale_names is not None:
      features |= {
          'rater_locale': tfds.features.ClassLabel(
              names=self.builder_config.rater_locale_names,
              doc='The locale of the rater.',
          )
      }

    if self.builder_config.includes_time_information:
      features |= {
          'answer_time_ms': tfds.features.Scalar(
              dtype=np.int64,
              doc=(
                  'Amount of time spent by each rater on each safety annotation'
                  ' question.'
              ),
          ),
          'answer_timestamp': tfds.features.Scalar(
              dtype=np.int64,
              doc='Time when each conversation was rated by each rater.',
          ),
      }

    if self.builder_config.safety_gold_names is not None:
      features |= {
          'safety_gold': tfds.features.ClassLabel(
              names=self.builder_config.safety_gold_names,
              doc='The gold standard safety label provided by experts.',
          ),
          'safety_gold_reason': tfds.features.Text(
              doc=(
                  'The reason(s) (if given) for the gold safety label provided'
                  ' by experts.'
              )
          ),
      }

    features |= {
        question: tfds.features.ClassLabel(names=['No', 'Yes', 'Unsure'])
        for question in self.builder_config.questions
    }

    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(features),
        supervised_keys=None,
        homepage='https://github.com/google-research-datasets/dices-dataset',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    name = self.builder_config.name
    filename = f'diverse_safety_adversarial_dialog_{name}.csv'
    paths = dl_manager.download_and_extract(
        {filename: _BASE_URL + f'{name}/{filename}'}
    )

    return {'train': self._generate_examples(paths[filename])}

  def _generate_examples(self, path):
    """Yields examples."""
    df = pd.read_csv(path).rename(columns={'rater_race_raw': 'rater_raw_race'})

    df['rater_raw_race'] = df['rater_raw_race'].fillna('')
    df['degree_of_harm'] = df['degree_of_harm'].fillna('No annotation')
    df['harm_type'] = (
        # NaN means there are no annotations
        df['harm_type']
        .fillna('No annotation')
        .str.split(',')
        # Standardize harm type names
        .map(lambda ts: [t.capitalize() for t in ts])
    )

    if self.builder_config.safety_gold_names is not None:

      def _is_parsable(s):
        try:
          ast.literal_eval(s)
        except (SyntaxError, ValueError):
          return False
        return True

      logging.info(
          'Fixing %d malformed safety_gold_reason strings',
          (~df['safety_gold_reason'].map(_is_parsable)).sum(),
      )

      df['safety_gold_reason'] = (
          df['safety_gold_reason']
          # Some safety gold reason features cannot be parsed by
          # `ast.literal_eval` because the string is malformed. To fix that, we
          # first pare the string down to a comma-separated sequence of safety
          # gold reason labels.
          .str.replace(r'[\[" \]]', '', regex=True)
          # We then recombine the sequence into a well-formed representation of
          # the list of safety gold reason string labels.
          .str.split(',').map(lambda lst: '["' + '","'.join(lst) + '"]')
      )

      # Complain loudly if the above fix does not work.
      if not df['safety_gold_reason'].map(_is_parsable).all():
        raise RuntimeError(
            'unable to fix all malformed safety_gold_reason strings'
        )

    for i, row in df.to_dict(orient='index').items():
      yield i, row



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""my_dataset dataset."""

from tensorflow_datasets.datasets.imagenet2012_real import imagenet2012_real_dataset_builder
import tensorflow_datasets.public_api as tfds


class ImageNet2012RealTest(tfds.testing.DatasetBuilderTestCase):
  """Test for ImageNet2012Real dataset."""

  DATASET_CLASS = imagenet2012_real_dataset_builder.Builder
  SPLITS = {"validation": 10}
  DL_DOWNLOAD_RESULT = "real.json"


if __name__ == "__main__":
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Version utils."""

from __future__ import annotations

import dataclasses
import enum
import re
from typing import Union

from etils import epath

_VERSION_TMPL = r"^(?P<major>{v})" r"\.(?P<minor>{v})" r"\.(?P<patch>{v})$"
_NO_LEADING_ZEROS = r"\d|[1-9]\d*"
_VERSION_WILDCARD_REG = re.compile(
    _VERSION_TMPL.format(v=_NO_LEADING_ZEROS + r"|\*")
)
_VERSION_RESOLVED_REG = re.compile(_VERSION_TMPL.format(v=_NO_LEADING_ZEROS))


class DatasetVariantBlockedError(ValueError):
  """Exception raised when a blocked version and/or config is requested."""


# A dictionary of blocked versions or configs.
# The key is a version or config string, the value is a short sentence
# explaining why that version or config should not be used (or None).
BlockedWithMsg = dict[str, str | None]
VersionOrStr = Union["Version", str]


@dataclasses.dataclass(frozen=True)
class IsBlocked:
  """Class to store information about a version or config being blocked.

  Also contains an optional message explaining why the version or config is
  blocked.
  """

  result: bool = False
  blocked_msg: str | None = None

  def __bool__(self) -> bool:
    return self.result


@dataclasses.dataclass(frozen=True)
class BlockedVersions:
  """Holds information on versions and configs that should not be used.

  Note that only complete versions can be blocked: wilcards in versions are not
  supported.

  versions: A dictionary of bad versions for which all configs should be
    blocked.
  configs: A mapping from versions to a dictionary of configs that should not be
    used for that version.
  """

  versions: BlockedWithMsg = dataclasses.field(default_factory=dict)
  configs: dict[str, BlockedWithMsg] = dataclasses.field(default_factory=dict)

  def is_blocked(
      self, version: VersionOrStr, config: str | None = None
  ) -> IsBlocked:
    """Checks whether a version or config is blocked.

    Args:
      version: The version to check.
      config: The config to check. If None, the version is checked.

    Returns:
      An IsBlocked object. If IsBlocked.result is True, IsBlocked.blocked_msg
      contains the message explaining why the version or config is blocked, if
      it exists, or a default message otherwise.
    """
    if isinstance(version, Version):
      version = str(version)
    if version in self.versions:
      blocked_msg = self.versions[version] or f"Version {version} is blocked."
      return IsBlocked(True, blocked_msg)
    if config is not None and version in self.configs:
      if config in self.configs[version]:
        blocked_msg = (
            self.configs[version][config]
            or f"Config {config} for version {version} is blocked."
        )
        return IsBlocked(True, blocked_msg)
    return IsBlocked(False)


class Experiment(enum.Enum):
  """Experiments which can be enabled/disabled on a per version basis.

  Experiments are designed to gradually apply changes to datasets while
  maintaining backward compatibility with previous versions. All experiments
  should eventually be deleted, once used by all versions of all datasets.

  Eg:
  class Experiment(enum.Enum):
    EXP_A = enum.auto()  # Short description of experiment.

  class MyBuilder(...):
    VERSION = tfds.core.Version('1.2.3', experiments={
        tfds.core.Experiment.EXP_A: True,
        })
  """

  # A Dummy experiment, which should NOT be used, except for testing.
  DUMMY = 1


class Version:
  """Dataset version MAJOR.MINOR.PATCH."""

  _DEFAULT_EXPERIMENTS = {
      Experiment.DUMMY: False,
  }

  def __init__(
      self,
      version: VersionOrStr,
      experiments=None,
      tfds_version_to_prepare=None,
  ):
    """Version init.

    Args:
      version: string. Eg: "1.2.3".
      experiments: dict of experiments. See Experiment.
      tfds_version_to_prepare: string, defaults to None. If set, indicates that
        current version of TFDS cannot be used to `download_and_prepare` the
        dataset, but that TFDS at version {tfds_version_to_prepare} should be
        used instead.
    """
    if isinstance(version, Version):
      version_str = str(version)
      experiments = experiments or version._experiments
      tfds_version_to_prepare = (
          tfds_version_to_prepare or version.tfds_version_to_prepare
      )
    else:
      version_str = version
    self._experiments = self._DEFAULT_EXPERIMENTS.copy()
    self.tfds_version_to_prepare = tfds_version_to_prepare
    if experiments:
      if isinstance(experiments, str):
        raise ValueError(
            f"Invalid Version('{version}', '{experiments}'). Description is "
            "deprecated. RELEASE_NOTES should be used instead."
        )
      self._experiments.update(experiments)
    self.major, self.minor, self.patch = _str_to_version(version_str)

  def implements(self, experiment):
    """Returns True if version implements given experiment."""
    return self._experiments[experiment]

  def __str__(self):
    return "{}.{}.{}".format(*self.tuple)

  def __repr__(self) -> str:
    return f"{type(self).__name__}('{str(self)}')"

  @property
  def tuple(self):
    return self.major, self.minor, self.patch

  def _validate_operand(self, other):
    if isinstance(other, str):
      return Version(other)
    elif isinstance(other, Version):
      return other
    raise AssertionError(
        "{} (type {}) cannot be compared to version.".format(other, type(other))
    )

  def __eq__(self, other):
    if other is None:
      return False
    other = self._validate_operand(other)
    return self.tuple == other.tuple

  def __lt__(self, other):
    other = self._validate_operand(other)
    return self.tuple < other.tuple

  def __le__(self, other):
    other = self._validate_operand(other)
    return self.tuple <= other.tuple

  def __gt__(self, other):
    other = self._validate_operand(other)
    return self.tuple > other.tuple

  def __ge__(self, other):
    other = self._validate_operand(other)
    return self.tuple >= other.tuple

  def __hash__(self) -> int:
    return hash(self.tuple)

  def match(self, other_version) -> bool:
    """Returns True if other_version matches.

    Args:
      other_version: string, of the form "x[.y[.x]]" where {x,y,z} can be a
        number or a wildcard.
    """
    major, minor, patch = _str_to_version(other_version, allow_wildcard=True)
    return (
        major in [self.major, "*"]
        and minor in [self.minor, "*"]
        and patch in [self.patch, "*"]
    )

  @classmethod
  def is_valid(cls, version: VersionOrStr | None) -> bool:
    """Returns True if the version can be parsed."""
    if isinstance(version, Version):
      return True
    elif version is None:
      return False
    try:
      return cls(version) and True
    except ValueError:  # Invalid version (ex: incomplete data dir)
      return False


def _str_to_version(
    version_str: str, allow_wildcard=False
) -> tuple[int | str, int | str, int | str]:
  """Return the tuple (major, minor, patch) version extracted from the str."""
  if not isinstance(version_str, str):
    raise TypeError(
        "Can only convert strings to versions. "
        f"Got: {type(version_str)} with value {version_str}."
    )
  reg = _VERSION_WILDCARD_REG if allow_wildcard else _VERSION_RESOLVED_REG
  res = reg.match(version_str)
  if not res:
    msg = "Invalid version '{}'. Format should be x.y.z".format(version_str)
    if allow_wildcard:
      msg += " with {x,y,z} being digits or wildcard."
    else:
      msg += " with {x,y,z} being numbers without leading zeros."
    raise ValueError(msg)
  return tuple(
      v if v == "*" else int(v)  # pylint:disable=g-complex-comprehension
      for v in [res.group("major"), res.group("minor"), res.group("patch")]
  )


def list_all_versions(root_dir: epath.PathLike) -> list[Version]:
  """Lists all dataset versions present on disk, sorted."""
  root_dir = epath.Path(root_dir)
  versions = []
  try:
    for path in root_dir.iterdir():
      if Version.is_valid(path.name) and path.is_dir():
        versions.append(Version(path.name))
  except OSError:
    return versions
  return sorted(versions)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to manually process RefCoco datasets.

Follow the instructions in the TFDS description.
"""

import json

from pycocotools.coco import COCO
from refer import REFER


def main():
  ref_data_root = '<path/to/refer/data/folder>'
  all_refs = []
  for dataset, split_bys in [
      ('refcoco', ['google', 'unc']),
      ('refcoco+', ['unc']),
      ('refcocog', ['google', 'umd']),
  ]:
    for split_by in split_bys:
      refer = REFER(ref_data_root, dataset, split_by)
      for ref_id in refer.getRefIds():
        ref = refer.Refs[ref_id]
        ann = refer.refToAnn[ref_id]
        ref['ann'] = ann
        ref['dataset'] = dataset
        ref['dataset_partition'] = split_by
        all_refs.append(ref)

  coco_annotations_file = '<path/to/instances_train2014.json>'
  coco = COCO(coco_annotations_file)
  ref_image_ids = set(x['image_id'] for x in all_refs)
  coco_anns = {
      image_id: {'info': coco.imgs[image_id], 'anns': coco.imgToAnns[image_id]}
      for image_id in ref_image_ids
  }

  out_file = '<path/to/refcoco.json>'
  with open(out_file, 'w') as f:
    json.dump({'ref': all_refs, 'coco_anns': coco_anns}, f)


if __name__ == '__main__':
  main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base TestCase to run inside context managers."""

from collections.abc import Sequence
import contextlib
from typing import Any, ContextManager

from absl.testing import absltest


class TestCaseInContext(absltest.TestCase):
  """Base TestCase for running tests inside the given contexts.

  It ensures that all contexts are entered before any test starts and are exited
  upon all tests completion.
  """

  @classmethod
  def _context_managers(cls) -> Sequence[ContextManager[Any]]:
    """Returns a list of context managers automatically applied to all tests."""
    raise NotImplementedError

  @classmethod
  def setUpClass(cls):
    super().setUpClass()

    # Keep track of all entered contexts
    cls._stack = contextlib.ExitStack()
    for cm in cls._context_managers():
      cls._stack.enter_context(cm)

  @classmethod
  def tearDownClass(cls):
    # Exit all contexts in the reverse order
    cls._stack.close()
    super().tearDownClass()

  def assertRaisesWithPredicateMatch(self, err_type, predicate):  # pylint: disable=invalid-name
    if isinstance(predicate, str):
      predicate_fn = lambda err: predicate in str(err)
    else:
      predicate_fn = predicate
    return super().assertRaisesWithPredicateMatch(err_type, predicate_fn)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CelebA dataset.

Large-scale CelebFaces Attributes (CelebA) Dataset

Deep Learning Face Attributes in the Wild
Ziwei Liu and Ping Luo and Xiaogang Wang and Xiaoou Tang
"""

from __future__ import annotations

import os
from typing import Sequence, Union

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

IMG_ALIGNED_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=0B7EVK8r0v71pZjFTYXZWM3FlRnM"
)
EVAL_LIST = (
    "https://drive.google.com/uc?export=download&"
    "id=0B7EVK8r0v71pY0NSMzRuSXJEVkk"
)
# Landmark coordinates: left_eye, right_eye etc.
LANDMARKS_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=0B7EVK8r0v71pd0FJY3Blby1HUTQ"
)

# Attributes in the image (Eyeglasses, Mustache etc).
ATTR_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=0B7EVK8r0v71pblRyaVFSWGxPY0U"
)

IDENTITY_DATA = (
    "https://drive.google.com/uc?export=download&"
    "id=1_ee_0u7vcNLOfNLegJRHmolfH5ICW-XS"
)

LANDMARK_HEADINGS = (
    "lefteye_x lefteye_y righteye_x righteye_y "
    "nose_x nose_y leftmouth_x leftmouth_y rightmouth_x "
    "rightmouth_y"
).split()
ATTR_HEADINGS = (
    "5_o_Clock_Shadow Arched_Eyebrows Attractive Bags_Under_Eyes Bald Bangs "
    "Big_Lips Big_Nose Black_Hair Blond_Hair Blurry Brown_Hair "
    "Bushy_Eyebrows Chubby Double_Chin Eyeglasses Goatee Gray_Hair "
    "Heavy_Makeup High_Cheekbones Male Mouth_Slightly_Open Mustache "
    "Narrow_Eyes No_Beard Oval_Face Pale_Skin Pointy_Nose Receding_Hairline "
    "Rosy_Cheeks Sideburns Smiling Straight_Hair Wavy_Hair Wearing_Earrings "
    "Wearing_Hat Wearing_Lipstick Wearing_Necklace Wearing_Necktie Young"
).split()

IDENTITY_HEADINGS = ("Identity_No").split()


class Builder(tfds.core.GeneratorBasedBuilder):
  """CelebA dataset. Aligned and cropped. With metadata."""

  VERSION = tfds.core.Version("2.1.0")
  RELEASE_NOTES = {
      "2.0.1": "New split API (https://tensorflow.org/datasets/splits)",
      "2.1.0": "Identity feature added.",
  }

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(
                shape=(218, 178, 3), encoding_format="jpeg"
            ),
            "landmarks": {name: np.int64 for name in LANDMARK_HEADINGS},
            # Attributes could be some special MultiLabel FeatureConnector
            "attributes": {name: np.bool_ for name in ATTR_HEADINGS},
            "identity": {name: np.int64 for name in IDENTITY_HEADINGS},
        }),
        homepage="http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html",
    )

  def _split_generators(self, dl_manager):
    downloaded_dirs = dl_manager.download({
        "img_align_celeba": IMG_ALIGNED_DATA,
        "list_eval_partition": EVAL_LIST,
        "list_attr_celeba": ATTR_DATA,
        "landmarks_celeba": LANDMARKS_DATA,
        "identity_celeba": IDENTITY_DATA,
    })

    # Load all images in memory (~1 GiB)
    # Use split to convert: `img_align_celeba/000005.jpg` -> `000005.jpg`
    all_images = {
        os.path.split(k)[-1]: img
        for k, img in dl_manager.iter_archive(
            downloaded_dirs["img_align_celeba"]
        )
    }

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "file_id": 0,
                "downloaded_dirs": downloaded_dirs,
                "downloaded_images": all_images,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "file_id": 1,
                "downloaded_dirs": downloaded_dirs,
                "downloaded_images": all_images,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "file_id": 2,
                "downloaded_dirs": downloaded_dirs,
                "downloaded_images": all_images,
            },
        ),
    ]

  def _process_celeba_config_file(self, file_path):
    """Unpack the celeba config file.

    The file starts with the number of lines, and a header.
    Afterwards, there is a configuration for each file: one per line.

    Args:
      file_path: Path to the file with the configuration.

    Returns:
      keys: names of the attributes
      values: map from the file name to the list of attribute values for
              this file.
    """
    with epath.Path(file_path).open() as f:
      data_raw = f.read()
    lines = data_raw.split("\n")

    keys = lines[1].strip().split()
    values = {}
    # Go over each line (skip the last one, as it is empty).
    for line in lines[2:-1]:
      row_values = line.strip().split()
      # Each row start with the 'file_name' and then space-separated values.
      values[row_values[0]] = [int(v) for v in row_values[1:]]
    return keys, values

  def _process_celeba_nonheading_config_file(
      self, file_path: Union[os.PathLike[str], str], keys_list: Sequence[str]
  ):
    """Unpack the celeba config file.

    The file starts with the number of lines, and a header.
    Afterwards, there is a configuration for each file: one per line.

    Args:
      file_path: Path to the file with the configuration.
      keys_list: List of names of the attributes.

    Returns:
      keys: names of the attributes
      values: map from the file name to the list of attribute values for
              this file.
    """
    with epath.Path(file_path).open() as f:
      data_raw = f.read()
    lines = data_raw.split("\n")

    keys = keys_list
    values = {}
    # Go over each line (skip the last one, as it is empty).
    for line in lines[0:-1]:
      row_values = line.strip().split()
      # Each row start with the 'file_name' and then space-separated values.
      values[row_values[0]] = [int(v) for v in row_values[1:]]
    return keys, values

  def _generate_examples(self, file_id, downloaded_dirs, downloaded_images):
    """Yields examples."""

    img_list_path = downloaded_dirs["list_eval_partition"]
    landmarks_path = downloaded_dirs["landmarks_celeba"]
    attr_path = downloaded_dirs["list_attr_celeba"]
    identity_path = downloaded_dirs["identity_celeba"]

    with epath.Path(img_list_path).open() as f:
      files = [
          line.split()[0]
          for line in f.readlines()
          if int(line.split()[1]) == file_id
      ]

    attributes = self._process_celeba_config_file(attr_path)
    landmarks = self._process_celeba_config_file(landmarks_path)
    identity = self._process_celeba_nonheading_config_file(
        identity_path, IDENTITY_HEADINGS
    )

    for file_name in sorted(files):
      record = {
          "image": downloaded_images[file_name],
          "landmarks": {
              k: v for k, v in zip(landmarks[0], landmarks[1][file_name])
          },
          "attributes": {
              # atributes value are either 1 or -1, so convert to bool
              k: v > 0
              for k, v in zip(attributes[0], attributes[1][file_name])
          },
          "identity": {
              k: v for k, v in zip(identity[0], identity[1][file_name])
          },
      }
      yield file_name, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""schema_guided_dialogue dataset."""

from __future__ import annotations

import json

import numpy as np
import tensorflow_datasets.public_api as tfds

_DATA_URL = (
    "https://github.com/google-research-datasets/dstc8-schema-guided-dialogue"
)


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for schema_guided_dialogue dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            # The previous system utterance if one exists.
            "utterances": tfds.features.Sequence(tfds.features.Text()),
            # TODO(arunchaganty): include frame and state annotations
            "first_speaker": tfds.features.ClassLabel(names=["USER", "SYSTEM"]),
            "metadata": {
                "services": tfds.features.Sequence(
                    {
                        "name": np.str_,
                        # TODO(arunchaganty): include service definitions
                    }
                )
            },
        }),
        supervised_keys=None,
        homepage=_DATA_URL,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        _DATA_URL + "/archive/refs/heads/master.zip"
    )

    return {
        "train": self._generate_examples(
            path / "dstc8-schema-guided-dialogue-master" / "train"
        ),
        "dev": self._generate_examples(
            path / "dstc8-schema-guided-dialogue-master" / "dev"
        ),
        "test": self._generate_examples(
            path / "dstc8-schema-guided-dialogue-master" / "test"
        ),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    # TODO(arunchaganty): load the services schemas from "schema.json".

    for f in path.glob("*.json"):
      if f.name.endswith("schema.json"):
        continue

      for datum in json.loads(f.read_text()):
        id_ = datum["dialogue_id"]
        utterances = [turn["utterance"] for turn in datum["turns"]]
        first_speaker = datum["turns"][0]["speaker"]

        yield id_, {
            "utterances": utterances,
            "first_speaker": first_speaker,
            # TODO(arunchaganty): include frame and state annotations
            "metadata": {
                # TODO(arunchaganty): include schema definitions
                "services": [
                    {
                        "name": service,
                    }
                    for service in datum["services"]
                ],
            },
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""covr dataset."""

import json

from etils import epath
import tensorflow_datasets.public_api as tfds


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for covr dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'utterance': tfds.features.Text(),
            'scenes': tfds.features.Sequence(
                feature=tfds.features.Text(),
            ),
            'properties': tfds.features.Sequence(
                feature=tfds.features.Text(),
            ),
            'pattern_name': tfds.features.Text(),
            'program': tfds.features.Text(),
            'label': tfds.features.Text(),
            'images': tfds.features.Sequence(
                feature=tfds.features.Image(),
            ),
        }),
        supervised_keys=None,
        homepage='https://covr-dataset.github.io/',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    extracted_dirs = dl_manager.download_and_extract({
        'covr_dir': (
            'https://drive.google.com/uc?export=download&'
            'id=10xlQ6isRdGX94BypoqN6klniGeqdLBJA'
        ),
        'imsitu_dir': (
            'https://s3.amazonaws.com/my89-frame-annotation'
            '/public/of500_images.tar'
        ),
        'vg1_dir': 'https://cs.stanford.edu/people/rak248/VG_100K_2/images.zip',
        'vg2_dir': (
            'https://cs.stanford.edu/people/rak248/VG_100K_2/images2.zip'
        ),
    })

    # Each name is the image file name without the ".jpg" extension, which is
    # also used as the scene id in COVR.
    image_path_by_scene_id: dict[str, epath.Path] = {}
    image_globs = [
        extracted_dirs['vg1_dir'].glob('*/*.jpg'),
        extracted_dirs['vg2_dir'].glob('*/*.jpg'),
        extracted_dirs['imsitu_dir'].glob('of500_images/*/*.jpg'),
    ]
    for image_glob in image_globs:
      for image_path in image_glob:
        name = image_path.stem
        image_path_by_scene_id[name] = image_path
    path = extracted_dirs['covr_dir']
    return {
        'train': self._generate_examples(
            path / 'train.jsonl', image_path_by_scene_id
        ),
        'test': self._generate_examples(
            path / 'test.jsonl', image_path_by_scene_id
        ),
        'validation': self._generate_examples(
            path / 'val.jsonl', image_path_by_scene_id
        ),
    }

  def _generate_examples(
      self, path: epath.Path, image_path_by_scene_id: dict[str, epath.Path]
  ):
    """Yields examples."""
    with path.open() as f:
      for line in f:
        item = json.loads(line)
        images = [
            image_path_by_scene_id[scene_id] for scene_id in item['scenes']
        ]
        yield item['qid'], {
            'utterance': item['utterance'],
            'scenes': item['scenes'],
            'properties': item['properties'],
            'pattern_name': item['pattern_name'],
            'program': str(item['program']),
            'label': str(item.get('answer')),
            'images': images,
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorFlow dataset for Binary Alphadigits."""
import urllib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URL = 'https://cs.nyu.edu/~roweis/data/'

_DESCRIPTION = (
    "Binary 20x16 digits of '0' through '9' and capital 'A' "
    "through 'Z'. 39 examples of each class."
)

_IMAGE_SHAPE = (20, 16, 1)

_NAMES = [
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    'M',
    'N',
    'O',
    'P',
    'Q',
    'R',
    'S',
    'T',
    'U',
    'V',
    'W',
    'X',
    'Y',
    'Z',
]


class Builder(tfds.core.GeneratorBasedBuilder):
  """Binary alphadigits dataset."""

  VERSION = tfds.core.Version('1.0.0')

  def _info(self):
    """Define the Dataset info."""

    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(shape=_IMAGE_SHAPE),
            'label': tfds.features.ClassLabel(names=_NAMES),
        }),
        supervised_keys=('image', 'label'),
        homepage=_URL,
    )

  def _split_generators(self, dl_manager):
    """Define Splits for training data."""

    path = dl_manager.download(
        {'train': urllib.parse.urljoin(_URL, 'binaryalphadigs.mat')}
    )

    return [
        tfds.core.SplitGenerator(
            name='train',
            gen_kwargs={
                'data_dir_path': path['train'],
            },
        )
    ]

  def _generate_examples(self, data_dir_path):
    """Generate Splits for training data."""

    with tf.io.gfile.GFile(data_dir_path, 'rb') as f:
      mat = tfds.core.lazy_imports.scipy.io.loadmat(f)
    for i in range(len(mat['dat'])):
      label = mat['classlabels'][0][i].item()
      for j in range(len(mat['dat'][i])):
        image = mat['dat'][i][j].reshape(20, 16, 1)
        yield '%d_%d' % (i, j), {'label': label, 'image': image}



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""wake_vision dataset."""

from collections.abc import Sequence
import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_TRAIN_IMAGE_IDS = [
    9270320,
    9270321,
    9270322,
    9270323,
    9270324,
    9270325,
    9270326,
    9270327,
    9270328,
    9270329,
    9270330,
    9270331,
    9270332,
    9270333,
    9270334,
    9270335,
    9270336,
    9270337,
    9270338,
    9270339,
    9270340,
    9270341,
    9270342,
    9270343,
    9270344,
    9270345,
    9270346,
    9270347,
    9270348,
    9270349,
    9270350,
    9270351,
    9270352,
    9270353,
    9270354,
    9270356,
    9270357,
    9270358,
    9270359,
    9270360,
    9270361,
    9270362,
    9270363,
    9270364,
    9270366,
    9270367,
    9270368,
    9270369,
    9270370,
    9270371,
    9270372,
    9270373,
    9270374,
    9270375,
    9270376,
    9270377,
    9270378,
    9270379,
    9270380,
    9270381,
    9270382,
    9270383,
    9270384,
    9270385,
    9270386,
    9270387,
    9270388,
    9270390,
    9270391,
    9270392,
    9270393,
    9270394,
    9270395,
    9270396,
    9270397,
    9270398,
    9270399,
    9270400,
    9270401,
    9270402,
    9270403,
    9270404,
    9270405,
    9270406,
    9270407,
    9270408,
    9270409,
    9270410,
    9270411,
    9270412,
]
_URL_TEMPLATE = (
    'https://dataverse.harvard.edu/api/access/datafile/{id}?format=original'
)
_IMAGE_URLS = {
    'train_images': [
        tfds.download.Resource(
            url=_URL_TEMPLATE.format(id=id),
            extract_method=tfds.download.ExtractMethod.GZIP,
        )
        for id in _TRAIN_IMAGE_IDS
    ],
    'validation_images': [
        tfds.download.Resource(
            url=_URL_TEMPLATE.format(id=9270355),
            extract_method=tfds.download.ExtractMethod.GZIP,
        )
    ],
    'test_images': [
        tfds.download.Resource(
            url=_URL_TEMPLATE.format(id=9270389),
            extract_method=tfds.download.ExtractMethod.GZIP,
        )
    ],
}
_METADATA_URLS = {
    'train_large_metadata': _URL_TEMPLATE.format(id=9844933),
    'train_quality_metadata': _URL_TEMPLATE.format(id=9844934),
    'validation_metadata': _URL_TEMPLATE.format(id=10243658),
    'test_metadata': _URL_TEMPLATE.format(id=10243659),
}

_DESCRIPTION = (
    'Wake Vision is a large, high-quality dataset featuring over 6 million'
    ' images, significantly exceeding the scale and diversity of current tinyML'
    ' datasets (100x). This dataset includes images with annotations of whether'
    ' each image contains aperson. Additionally, it incorporates a'
    ' comprehensive fine-grained benchmark to assess fairness and robustness,'
    ' covering perceived gender, perceived age, subject distance, lighting'
    ' conditions, and depictions. The Wake Vision labels are derived from Open'
    " Image's annotations which are licensed by Google LLC under CC BY 4.0"
    ' license. The images are listed as having a CC BY 2.0 license. Note from'
    ' Open Images: "while we tried to identify images that are licensed under a'
    ' Creative Commons Attribution license, we make no representations or'
    ' warranties regarding the license status of each image and you should'
    ' verify the license for each image yourself."'
)


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for wake_vision dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': (
          'Initial TensorFlow Datasets release. Note that this is based on the'
          ' 2.0 version of Wake Vision on Harvard Dataverse.'
      ),
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            # These are the features of your dataset like images, labels ...
            'image': tfds.features.Image(shape=(None, None, 3)),
            'filename': tfds.features.Text(),
            'person': tfds.features.ClassLabel(names=['No', 'Yes']),
            'depiction': tfds.features.ClassLabel(names=['No', 'Yes']),
            'body_part': tfds.features.ClassLabel(names=['No', 'Yes']),
            'predominantly_female': tfds.features.ClassLabel(
                names=['No', 'Yes']
            ),
            'predominantly_male': tfds.features.ClassLabel(names=['No', 'Yes']),
            'gender_unknown': tfds.features.ClassLabel(names=['No', 'Yes']),
            'young': tfds.features.ClassLabel(names=['No', 'Yes']),
            'middle_age': tfds.features.ClassLabel(names=['No', 'Yes']),
            'older': tfds.features.ClassLabel(names=['No', 'Yes']),
            'age_unknown': tfds.features.ClassLabel(names=['No', 'Yes']),
            'near': tfds.features.ClassLabel(names=['No', 'Yes']),
            'medium_distance': tfds.features.ClassLabel(names=['No', 'Yes']),
            'far': tfds.features.ClassLabel(names=['No', 'Yes']),
            'dark': tfds.features.ClassLabel(names=['No', 'Yes']),
            'normal_lighting': tfds.features.ClassLabel(names=['No', 'Yes']),
            'bright': tfds.features.ClassLabel(names=['No', 'Yes']),
            'person_depiction': tfds.features.ClassLabel(names=['No', 'Yes']),
            'non-person_depiction': tfds.features.ClassLabel(
                names=['No', 'Yes']
            ),
            'non-person_non-depiction': tfds.features.ClassLabel(
                names=['No', 'Yes']
            ),
        }),
        supervised_keys=('image', 'person'),
        homepage='https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi%3A10.7910%2FDVN%2F1HOPXC',
        license='See homepage for license information.',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    metadata_paths = dl_manager.download_and_extract(_METADATA_URLS)
    image_paths = dl_manager.download_and_extract(_IMAGE_URLS)

    return {
        'train_large': self._generate_examples(
            image_paths['train_images'], metadata_paths['train_large_metadata']
        ),
        'train_quality': self._generate_examples(
            image_paths['train_images'],
            metadata_paths['train_quality_metadata'],
        ),
        'validation': self._generate_examples(
            image_paths['validation_images'],
            metadata_paths['validation_metadata'],
        ),
        'test': self._generate_examples(
            image_paths['test_images'], metadata_paths['test_metadata']
        ),
    }

  def _generate_examples(
      self, image_paths: Sequence[epath.Path], metadata_path: epath.Path
  ):
    """Yields examples."""
    metadata = tfds.core.lazy_imports.pandas.read_csv(
        metadata_path, index_col='filename'
    )

    for tar_file in image_paths:
      for sample_path, sample_object in tfds.download.iter_archive(
          tar_file, tfds.download.ExtractMethod.TAR_STREAM
      ):
        file_name = os.path.basename(sample_path)

        if file_name not in metadata.index:
          continue

        sample_metadata = metadata.loc[file_name]

        yield file_name, {
            'image': sample_object,
            'filename': file_name,
            'person': sample_metadata['person'],
            'depiction': sample_metadata.get('depiction', -1),
            'body_part': sample_metadata.get('body_part', -1),
            'predominantly_female': sample_metadata.get(
                'predominantly_female', -1
            ),
            'predominantly_male': sample_metadata.get('predominantly_male', -1),
            'gender_unknown': sample_metadata.get('gender_unknown', -1),
            'young': sample_metadata.get('young', -1),
            'middle_age': sample_metadata.get('middle_age', -1),
            'older': sample_metadata.get('older', -1),
            'age_unknown': sample_metadata.get('age_unknown', -1),
            'near': sample_metadata.get('near', -1),
            'medium_distance': sample_metadata.get('medium_distance', -1),
            'far': sample_metadata.get('far', -1),
            'dark': sample_metadata.get('dark', -1),
            'normal_lighting': sample_metadata.get('normal_lighting', -1),
            'bright': sample_metadata.get('bright', -1),
            'person_depiction': sample_metadata.get('person_depiction', -1),
            'non-person_depiction': sample_metadata.get(
                'non-person_depiction', -1
            ),
            'non-person_non-depiction': sample_metadata.get(
                'non-person_non-depiction', -1
            ),
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SummScreen Summarization dataset, non-anonymized, non-tokenized version."""

import json
import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DL_URLS = {
    # pylint: disable=line-too-long
    'tokenized': 'https://drive.google.com/uc?export=download&id=1BvdIllGBo9d2-bzXQRzWuJXB04XPVmfF',
    'untokenized': 'https://drive.google.com/uc?export=download&id=1tFpt32USOO2i1FWhtFTsyYyFzuRm2k36',
    # pylint: enable=line-too-long
}

_RECAP = 'recap'
_TRANSCRIPT = 'transcript'
_RECAP_SOURCE_FULL_NAMES = {
    'fd': 'ForeverDreaming',
    'tms': 'TVMegaSite',
}
_SPLITS = ['train', 'dev', 'test']


def _load_file(path):
  with tf.io.gfile.GFile(path, 'r') as f:
    return f.read()


def _load_json(path):
  return json.loads(_load_file(path))


def _load_jsonl(path):
  return [json.loads(line) for line in _load_file(path).strip().splitlines()]


def _get_filenames_dict(tokenized_path, recap_source: str):
  """Get dictionary of filenames for each split."""
  filenames_dict = {}
  for split in _SPLITS:
    tokenized_data = _load_jsonl(
        os.path.join(
            tokenized_path,
            'SummScreen',
            _RECAP_SOURCE_FULL_NAMES[recap_source],
            f'{recap_source}_{split}.json',
        )
    )
    filenames_dict[split] = [row['filename'] for row in tokenized_data]
  return filenames_dict


def _get_paths_dict(untokenized_path, recap_source, filenames_dict):
  """Get dictionary of example paths for each split."""
  paths_dict = {}
  for split, filenames in filenames_dict.items():
    paths_dict[split] = [
        os.path.join(untokenized_path, 'SummScreen_raw', recap_source, filename)
        for filename in filenames
    ]
  return paths_dict


class SummscreenConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Summscreen."""

  def __init__(self, *, recap_source=None, **kwargs):
    """BuilderConfig for Summscreen.

    Args:
      recap_source: str. The directory for the source of recaps to read.
      **kwargs: keyword arguments forwarded to super.
    """
    super(SummscreenConfig, self).__init__(**kwargs)
    self.recap_source = recap_source


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for non-tokenized, non-anonymized SummScreen dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  BUILDER_CONFIGS = [
      SummscreenConfig(
          name='fd',
          description='ForeverDreaming',
          recap_source='fd',
      ),
      SummscreenConfig(
          name='tms',
          description='TVMegaSite',
          recap_source='tms',
      ),
  ]

  def _info(self):
    # Should return a tfds.core.DatasetInfo object
    if self._builder_config.recap_source == 'fd':  # pytype: disable=attribute-error  # always-use-return-annotations
      features = tfds.features.FeaturesDict({
          _TRANSCRIPT: tfds.features.Text(),
          _RECAP: tfds.features.Text(),
          'episode_number': tfds.features.Text(),
          'episode_title': tfds.features.Text(),
          'show_title': tfds.features.Text(),
          'transcript_author': tfds.features.Text(),
      })
    elif self._builder_config.recap_source == 'tms':  # pytype: disable=attribute-error  # always-use-return-annotations
      features = tfds.features.FeaturesDict({
          _TRANSCRIPT: tfds.features.Text(),
          _RECAP: tfds.features.Text(),
          'episode_summary': tfds.features.Text(),
          'show_title': tfds.features.Text(),
          'transcript_author': tfds.features.Tensor(
              shape=(None,), dtype=np.str_
          ),
          'recap_author': tfds.features.Text(),
      })
    else:
      raise KeyError(
          f'Unknown recap_source {self._builder_config.recap_source}'  # pytype: disable=attribute-error  # always-use-return-annotations
      )

    return self.dataset_info_from_configs(
        features=features,
        supervised_keys=(_TRANSCRIPT, _RECAP),
        homepage='https://github.com/mingdachen/SummScreen',
    )

  def _split_generators(self, dl_manager):
    dl_paths = dl_manager.download_and_extract(_DL_URLS)
    filenames_dict = _get_filenames_dict(
        tokenized_path=dl_paths['tokenized'],
        recap_source=self._builder_config.recap_source,  # pytype: disable=attribute-error  # always-use-return-annotations
    )
    paths_dict = _get_paths_dict(
        untokenized_path=dl_paths['untokenized'],
        recap_source=self._builder_config.recap_source,  # pytype: disable=attribute-error  # always-use-return-annotations
        filenames_dict=filenames_dict,
    )
    return {
        'train': self._generate_examples(paths=paths_dict['train']),
        'validation': self._generate_examples(paths=paths_dict['dev']),
        'test': self._generate_examples(paths=paths_dict['test']),
    }

  def _generate_examples(self, paths):
    for path in paths:
      example = _load_json(path)
      fname = os.path.basename(path)
      if self._builder_config.recap_source == 'fd':  # pytype: disable=attribute-error  # always-use-return-annotations
        yield fname, {
            _TRANSCRIPT: '\n'.join(example['Transcript']),
            _RECAP: '\n'.join(example['Recap']),
            'episode_number': example['Episode Number'],
            'episode_title': example['Episode Title'],
            'show_title': example['Show Title'],
            'transcript_author': example['Transcript Author'],
        }
      elif self._builder_config.recap_source == 'tms':  # pytype: disable=attribute-error  # always-use-return-annotations
        yield fname, {
            _TRANSCRIPT: '\n'.join(example['Transcript']),
            _RECAP: '\n'.join(example['Recap']),
            'episode_summary': '\n'.join(example['Episode Summary']),
            'show_title': example['Show Title'],
            'transcript_author': example['Transcript Author'],
            'recap_author': example['Recap Author'],
        }
      else:
        raise KeyError(
            f'Unknown recap_source {self._builder_config.recap_source}'  # pytype: disable=attribute-error  # always-use-return-annotations
        )
